-- Generalized Railway Crossing example
-- 
--verify both the safety and utility requirement
 
var 
	x1, 	-- 	for train 1, measure distance from I 
	 	--	values stored are negative
	x2: clock; 
	g, 	--  for gate angle, ranges from 0=down to 90=up 
	y1,y2:analog; -- clocks used for utility property




automaton train1
initially far & x1<=0 & y1=0;
synclabs: enterR1,  exit1;

loc far: while x1 <= 0 wait {dy1=1}
	when x1=0  sync enterR1 goto S;

loc S: while x1 <= 40 wait {dy1=1}
	when 30<=x1 & x1 <= 40 goto I;

loc I: while True wait {dy1=1}
	when True do {x1' <= 0,y1' = 0} sync exit1 goto far;
end

automaton train2
initially far & x2<=0 & y2=0;
synclabs: enterR2, exit2;

loc far: while x2 <= 0 wait {dy2=1}
	when x2=0 sync enterR2 goto S;

loc S: while x2 <= 40 wait {dy2=1}
	when 30<=x2 & x2<=40  goto I;

loc I: while True wait {dy2=1}
	when True do {x2' <=0,y2' = 0} sync exit2 goto far;
end


automaton gate
initially g_up & g=90;
synclabs: lower, raise;

loc g_up: while True wait {dg=0}
	when True sync lower goto g_lowering;
	when True sync raise goto g_up;

loc g_down: while True wait {dg=0}
	when True sync lower goto g_down;
	when True sync raise goto g_raising;


loc g_raising: while g<=90 wait {dg in [9,10]}
	when True sync raise goto g_raising;
	when True sync lower goto g_lowering;
	when g=90 goto g_up;

loc g_lowering: while g>=0 wait {dg in [-10,-9]}
	when True sync raise goto g_raising;
	when True sync lower goto g_lowering;
	when g=0 goto g_down;

end


automaton controller
initially up;
synclabs: exit1, exit2, raise, lower, enterR1, enterR2;

loc up: while x1<=19 & x2 <= 19 wait {}
	when True sync exit1 goto up;
	when True sync exit2 goto up;
	when True sync enterR1 goto up;
	when True sync enterR2 goto up;
	when x1=19 sync lower goto down;
	when x2=19 sync lower goto down;

loc down: while True wait {}
	when x1<=19 sync exit2 goto c_must_raise;
	when x2<=19 sync exit1 goto c_must_raise;
	when 19<=x1 sync exit2 goto down;
	when 19<=x2 sync exit1 goto down;

	when True sync enterR1  goto down;
	when True sync enterR2  goto down;


loc c_must_raise: while True wait {}
	when asap sync raise goto up;	
end



-- analysis commands

var 
   init_reg, final_reg : region;

init_reg := loc[train1] = far & x1<=0 &
	    loc[train2] = far & x2<=0 &
 	    loc[gate] = g_up & g=90 &
            loc[controller] = up;

final_reg := 	 
	loc[train1]=I &
	(loc[gate]=g_up | loc[gate]=g_raising| loc[gate]=g_lowering) 
	|
	(loc[train1] = far | loc[train1]= S) & 
	(loc[train2]=far | loc[train2]=S) &
	(loc[gate] = g_down | loc[gate] = g_raising | loc[gate]=g_lowering) & 
	x1<=18 & x2<=18 & y1>=11 & y2>=11;
 

if empty(reach forward from init_reg endreach & final_reg)
   then prints "GRC satisfies safety and utility requirements";
   else prints "GRC does not satisfy safety and utility requirements";
endif;


