-- dCorbetts distributed controller

-- synthesizing maximal delay time between control signals

var
	y1,		-- stopwatch for sensor_1
	y2,		-- stopwatch for sensor_2
	x1,      	-- dschedulers stopwatch for delay on sensor 1 
	x2,      	-- dschedulers stopwatch for delay on sensor 2
	z		-- dcontrollers stopwatch
	: stopwatch;
	c		-- to measure maximal time delay?
	: clock;
	alpha
	: parameter;

 -- -------------------------------------------------------------- *)

automaton sensor_1
synclabs : request_1, read_1, send_1, ack_1;

initially done & y1=6;

loc done: while y1<=6 wait {dy1 = 1}
	when y1>=6 sync request_1 goto read;
	
loc read: while True wait {dy1 = 0}
	when True sync read_1 do {y1'=0} goto waiting;

loc waiting: while y1<=4 wait {dy1=1}
	when y1>=4 sync request_1 goto read;
	when asap sync send_1 goto send;

loc send: while True wait {dy1=0}
	when True do {y1' = 0} sync ack_1 goto done;
end


-- -------------------------------------------------

automaton sensor_2
synclabs : request_2, read_2, send_2, ack_2;

initially done & y2=6;

loc done: while y2<=6 wait {dy2 = 1}
	when y2>=6 sync request_2 goto read;
	
loc read: while True wait {dy2 = 0}
	when True do {y2' =0} sync read_2 goto waiting;

loc waiting: while y2<=8 wait {dy2=1}
	when y2>=8 sync request_2 goto read;
	when asap sync send_2 goto send;

loc send: while True wait {dy2=0}
	when True do {y2' = 0} sync ack_2 goto done;
end


-- -------------------------------------------------



automaton scheduler
synclabs : request_1, read_1, request_2, read_2;

initially idle;

loc idle: while True wait {dx1=0,dx2=0}
	when True  sync request_1 do {x1' = 0} goto sensor1;
	when True  sync request_2 do {x2' = 0} goto sensor2;
	
loc sensor1: while x1<=11/10 wait {dx1=1,dx2=0}
	when x1>=1/2 sync read_1 goto idle;
	when True sync request_2 do {x2'=0} goto sensor2_wait1;

loc sensor2: while x2<=2 wait {dx2=1,dx1=0}
	when x2>=3/2 sync read_2 goto idle;
	when True sync request_1 do {x1'=0} goto sensor2_wait1;

loc sensor2_wait1: while x2<=2 wait {dx2=1,dx1=0}
	when x2>=3/2 sync read_2 goto sensor1;
	

end


automaton controller
synclabs : 	send_1, expire_1, ack_1, 
		send_2, expire_2, ack_2, 
		signal;

initially rest & c=0;

loc rest: while True wait {dz=0}
	when True sync send_1 do {z'=0} goto rec1_1;
	when True sync send_2 do {z'=0} goto rec2_2;

loc rec1_1: while z<=1 wait {dz=1}
	when z>=9/10 sync ack_1 goto wait2;

loc wait2: while z<=10 wait {dz=1}
	when z>=10 sync expire_2 goto rest;
	when True sync send_2 do {z'=0} goto rec1_2;

loc rec1_2: while z<=1 wait {dz=1}
	when z>=9/10 sync ack_2 do {z'=0} goto compute;

loc rec2_2: while z<=1 wait {dz=1}
	when z>=9/10 sync ack_2 goto wait1;

loc wait1: while z<=10 wait {dz=1}
	when z>=10 sync expire_1 goto rest;
	when True sync send_1 do {z'=0} goto rec2_1;

loc rec2_1: while z<=1 wait {dz=1}
	when z>=9/10 sync ack_1 do {z'=0} goto compute;

loc compute: while z<=56/10 wait {dz=1}
	when z>=36/10 sync signal do {c'=0} goto rest;

end

prints "Alpha is possible delay time between control signals";
prints "Following condition shows all achievable values of alpha";

print omit all locations 
	hide non_parameters in 
		reach forward from 
			loc[sensor_1] = done & y1=6 &
			loc[sensor_2] = done & y2=6 &
			loc[scheduler] = idle &
			loc[controller] = rest & c=0
		endreach
		& c >= alpha
	endhide;
