// program to make C++ automaton from C structure automaton
//
// also provide service to Pre-post code, by accessing the automaton.
//
// make the product in a separate pair of files product.c, product.H
//

// notes:
// check on initialization, and constructors
//

// #include <strings.h>
#include "poly.H"
#include "extrapoly.H"
#include "automaton.H"


/* =============== global vars ==================== */

// have a default vector in 1st place, for manipulating list.
slope_vectors_info the_slopes;

automaton global_a;

/* =============== misc ==================== */


char * TRUEstring = "TRUE";
char * FALSEstring = "FALSE";
char *varnames[MAX_VARS];  // names of all the variables
Bool elim_var_index[MAX_VARS];
Bool elim_proc_index[MAX_PROCS];
char *infilename;


// options/flags
int print_mode = 0;
int format_mode = 0;
Bool check_automaton_flag = true;
Bool reach_with_edges = true;


// globals for reachability
Bool did_reach_with_edges = false;
int max_reg_id = -1;
Bool generate_forward_trace = true;  // absolute direction of trace
Bool reached_forward = true;
Bool generating_trace = false;


// global for debugging
int debug_magic_nr;


void
put_thin_line() {
 cout << "----------------------------\n";
}

void
put_thick_line() {
 cout << "===========================\n";
}

void
put_line() {
  put_thin_line();
}

char *
Bool_output(Bool b) {
  if (b) {
    return(TRUEstring);
  }
  else {
    return(FALSEstring);
  }
}

void
swap_by_addr(int *p1, int *p2) {
  int temp = *p1;
  *p1 = *p2;
  *p2 = temp;
}

void
global_free_str(char *str) {
  if (str != NULL)
    free(str);
}

void
global_free(void *p) {
  if (p!=NULL) 
    free(p);
}

void
trace_out(char *outputstring, coeff num) {
  cout << outputstring << num << "\n" << flush;
}


coeff lcm_coeff(coeff x, coeff y) {
  coeff tmp;
  Gcd(x, y, &tmp);
  return x * (y / tmp);
}


void
clear_ptr_int_array(int **arr, int nr_elts){
   int i;
   for (i=0;i<nr_elts;i++) {
     arr[i] = (int *) NULL;
   }
}


coeff 
abs_val(coeff a) {  
return( (a>=0) ? a : -a );
}

coeff min_of(coeff a, coeff b) {
  return( (a<b)?a:b );
}


coeff max_of(coeff a, coeff b) {
  return( (a>b)?a:b );
}

// return index of posn of key in the int array arr.
// return -1 if not in the array
int  
pos_in_int_array(int key, int *arr, int size) {
   Bool found = false;
   int i=0;
   while (!found && (i<size)) {
     found = (key == arr[i]);
     if (!found) {
       i++;
     }
   }
   if (found) {
     return (i);
   }
   else {
     return(-1);
   }
 }


void 
automaton::set_elim_info() {
  if (elim_locs_by_list) {
    int i;

    struct var_name *x = elim_proc_list;
    for (i=0;i<MAX_PROCS;i++) {
      elim_proc_index[i] = false;
    }
    while (x != NULL) {
      Bool found = false;
      int aut_pos = 0;
      while  (!found && (aut_pos <nr_components)) {
	found = (strcmp(x->name,
			comp_inf_p[aut_pos].automaton_name) == 0);
	if (!found) {
	  aut_pos++;
	}
      }
      if (!found) {
	cout << "Error: bogus automaton name -- " << x->name << "\n";
	exit(1);
      }
      elim_proc_index[aut_pos] = true;
      x = x->next;
    }
  }
  if (elim_vars_by_list) {
    int i;
    struct var_name *x = elim_var_list;
    for (i=0;i<MAX_VARS;i++) {
      elim_var_index[i] = false;
    }
    while (x != NULL) {
      int var_pos = var_index(x->name);
      if (var_pos == -1) {
	cout << "Error: bogus variable name -- " << x->name << "\n";
	exit(1);
      }
      else {
	elim_var_index[var_pos] = true;
      }
      x = x->next;
    }
    for (i=0;i<MAX_VARS;i++) {
      if (elim_var_index[i]) {
	cout << "Registered need to elim " << i << "\n" << flush;
      }
    }
  }
}

/* ================ lists in general, see Stroustrup p266 ================= */

void
slist_base::insert(slink * a)  // add to head of list
{
  if (last)
    a->next = last->next;
  else
    last = a;
  last->next = a;
}

void
slist_handler(char *c) {
  cout << c << "\n" << "aborting\n" << flush;
  exit(2);
}


slink* slist_base::get() // return and remove head
{
  if (last==0) slist_handler("get from empty slist");
  slink* f = last->next;
  if (f==last)
    last = 0;
  else
    last->next = f->next;
  return f;
}


slist_base_iter::slist_base_iter(slist_base& s)
{
 cs = &s;
  ce = cs->last;
}

/* start here */




/* ================  matrix lists ============================= */



matrix_info::matrix_info(const matrix &new_m) {
  m = new_m;
  next = NULL;
}

matrix_info::matrix_info(int nb_cols) {
  int j;
  m = matrix(1,nb_cols);
  m.p[0][0] = 1;
  for (j=1;j<nb_cols-1;j++) {
    m.p[0][j]=0;
  }
  m.p[0][nb_cols-1] = 1;
  next = NULL;
}


matrix_info *
matrix_info::insert(matrix_info *mip) {
  // insert into head of singly-linked list 
  mip->next = this;
  return(mip);
}


void
matrix_info::print() {
// 
// recall. matrix_info used to store the actual single global list of 
// matrices.
  m.print();
  if (next != NULL) {
    next->print();
  }
}


/* ================  transitions (+ lists) ============================= */

transition::transition() {
  src = NULL;
  dest = NULL;
  sync_label = NULL;
  enabling_constr = NULL;
  is_asap_trans = false;
  ass = assignment();
}

transition::transition(location *new_src, location *new_dest, char *sync_lab,
		       matrix *gd, Bool is_asap, assignment *assgt) {
  src = new_src;
  dest = new_dest;
  is_asap_trans = is_asap;
  if (sync_lab) {
    sync_label = str_preserve(sync_lab);
  }
  else {
    // no sync label
    sync_label = NULL;
  }
  enabling_constr = gd;
  ass = *assgt;
}


void
transition::print() {
  cout << "trans " << src->name << " (" << src->locn_id << ") -> "
    << dest->name << " (" << dest->locn_id << ") on " << sync_label << "\n";
  cout << "ASAP: " << Bool_output(is_asap_trans) << "\n";
  cout << "with guard:\n\t";
  enabling_constr->print();
  cout << "Assignment\n";
  ass.print();
}

void
transition::brief_print() {
  cout << "trans " << src->name << " (" << src->locn_id << ") -> "
    << dest->name << " (" << dest->locn_id << ") on " << sync_label << "\n";
  cout << "ASAP: " << Bool_output(is_asap_trans) << "\n";
  cout << "with guard:\n\t";
  enabling_constr->ptprint(NFORM);
}


void
transition::label_print() {
  cout << "-------------------------------\n";
  if (sync_label != NULL) {
    cout << "  VIA: " <<  sync_label << "\n";
  }
  else {
    cout << "  VIA:\n";
  }
  cout << "-------------------------------\n";
}


void
transition::free() {
  if (sync_label != NULL) {
    global_free(sync_label);
  }
  ass.free();
}


transition_info::transition_info() {
  trans = transition();
  next = NULL;
}


transition_info::transition_info(transition &tr) {
// is default () constructor first, thereby setting next to NULL??
  trans = tr;
  next = NULL;
}


transition_info *
transition_info::insert(transition_info *tip) {
  // insert into head of singly-linked list 
  tip->next = this;
  return(tip);
}


/* ===============  assignments (+lists) =============== */


assignment::assignment() {
  nr_vars = 0;
  indices = NULL;
  nr_new_vars = 0;
  new_indices = NULL;
  ass_constraints = NULL;
}


assignment::assignment(int nbvars, int *ind, matrix *mp) {
// copy the indices
  int * p, *q;
  int i;

  nr_vars = nbvars;
  indices = (int *) malloc (nbvars * sizeof(int));
  p = indices; q = ind;
  for (i=0;i<nbvars;i++) {
    *p++ = *q++;
  }
  nr_new_vars = 0;
  new_indices = NULL;
  ass_constraints = mp;
}


void
assignment::free() {
  global_free(indices);
}


void
assignment::print() {
  int i;
  cout << "New vars (" << nr_vars << ") : ";
  for (i=0;i<nr_vars;i++) {
    cout << indices[i] << " ";
  }
  cout << "\nNew vars not used in old form (" << nr_new_vars << ") :";
  for (i=0;i<nr_new_vars;i++) {
    cout << new_indices[i] << " ";
  }
  cout  << "\n\t with added constraints:\n";
  ass_constraints->print();
}

void 
assignment::operator =(const assignment& a) {
  nr_vars = a.nr_vars;
  indices = a.indices;
  nr_new_vars  = a.nr_new_vars;
  new_indices = a.new_indices;
  ass_constraints = a.ass_constraints;
}

void
assignment::set_matrix_p(matrix *mp) {
  ass_constraints = mp;
}

// oldmp: ptr to old ass matrix
// need to set up *nr_old_vars_used_p as nr of old vars that 
//        get used in ass constraints
// set oldvarindicesp so that i'th value has var ID of i'th old var 
//        used in constraints.
// BUG fixed: 10/14/95.
// no, need the varID +1, i.e the exact column position.
// assume dealing with an nplus1_form, with var_count including the slack var
void
get_info_re_old_ass(matrix *oldmp,
		    int *nr_old_vars_used_p,
		    int **oldvarindicesp,
		    int *nr_new_vars_used_p,
		    int **newvarindicesp,
		    int *indices,
		    int nr_v) {
  int i,j;
  Bool used;
  int nr_r = oldmp->nbrows;
  *nr_old_vars_used_p = 0;
  *oldvarindicesp = NULL;

  *nr_new_vars_used_p = 0;
  *newvarindicesp = NULL;

  for (j=0;j<var_count-1;j++) {
    // for each var, check if it's used in old form of var AND is reassigned
    //    if so add to index list and inc count.
    // First: check if old form is used
    used = false;
    for (i=0;i<nr_r-1;i++) {
      if (oldmp->p[i][j+1] != 0) {
	used = true;
      }
    }

    if (used) {
      // Second: check  whether it is reassigned (i.e if primed form parsed in)
      //   Note: cannot use column entries, since y'=y' doesn't show up
      used = false;
      for (i=0;i<nr_v;i++) {
        if (indices[i]==j+1) {
	  used = true;
        }
      }
      if (used) {
	if (*nr_old_vars_used_p == 0) {
	  *oldvarindicesp = (int *) malloc (sizeof(int));
	}
	else {
	  *oldvarindicesp = 
	    (int *) realloc (*oldvarindicesp,
			     (*nr_old_vars_used_p+1)*sizeof(int));
	}
	(*oldvarindicesp)[*nr_old_vars_used_p] = j+1;
	(*nr_old_vars_used_p)++;
      }
    }
  }
  
  // now add info for vars that are reassigned, but not used in old form
  for (i=0;i<nr_v;i++) {
    int v_ind = indices[i];
    used = false;
    for (j=0;j<nr_r;j++) {
      if (oldmp->p[j][v_ind] != 0) {
	used = true;
      }
    }
    if (!used) {
	if (*nr_new_vars_used_p == 0) {
	  *newvarindicesp = (int *) malloc (sizeof(int));
	}
	else {
	  *newvarindicesp = 
	    (int *) realloc (*newvarindicesp,
			     (*nr_new_vars_used_p+1)*sizeof(int));
	}
	(*newvarindicesp)[*nr_new_vars_used_p] = v_ind;
	(*nr_new_vars_used_p)++;
      }
  }

}

assignment
assignment::make_inverse_ass() {
//
// given an assignment, return its inverse
// need to make the inverse matrix on free store, temp. store pointer to it
// in ass_constraints.  
// ideally would like to add to automaton's list of all matrices.
// but for now do this in the calling routine.
// assume eps_slack variable is in the col with index var_count
  int i,j;
  int nr_r, base_col, nr_v;
  int nr_cols;               // number cols in new assignmt matrix
  int nr_old_vars_used;
  int nr_new_vars_used;

  int eps_col = var_count;

  int *oldvarindices = NULL; // list of the names (ids) of the old vars needed
  int *newvarindices = NULL; // list of the names (ids) of the old vars needed
  int src_pos, dest_pos;      /* col #'s for x' vars in oldm and newm */
  matrix *oldm = ass_constraints;
  
  get_info_re_old_ass(oldm,&nr_old_vars_used,  &oldvarindices, 
		      &nr_new_vars_used, &newvarindices,
		      indices, nr_vars);
  nr_r = (*oldm).nbrows;
  nr_v = nr_vars;
  nr_cols = var_count + nr_old_vars_used +2;
  matrix *newm = new matrix(nr_r,nr_cols);
  newm->blank();

  // copy columns for var_count-1 "real" variables
  // vars that are reassigned are later copied over
  for (i=0;i<nr_r-1;i++) {
    for (j=0;j<var_count;j++) {
      newm->p[i][j] = ass_constraints->p[i][j];
    }
  }

  base_col = var_count+1;  // vars start at 0
  // set up the x' columns in the first var_count col positions
  for (j=0;j<nr_v;j++) {
    src_pos = base_col + j;
    dest_pos = indices[j];  // indices start at 1 
    for (i=0;i<nr_r-1;i++) {   /* can leave last row as is */
      newm->p[i][dest_pos] = ass_constraints->p[i][src_pos];
    }
  }

  // copy the cols for the original (unprimed vars) that are also reassigned
  for (j=0;j<nr_old_vars_used;j++) {
    if (oldvarindices[j] >0) {
      // old var indices start at 1, altered 10/14/95
      src_pos = oldvarindices[j]; 
      dest_pos = var_count+1+j;
      for (i=0;i<nr_r-1;i++) {
	newm->p[i][dest_pos] = ass_constraints->p[i][src_pos];
      }
    }
  }

  // copy the constants from last col, and copy the eps_slack variable
  for (i=0;i<nr_r-1;i++) {
    newm->p[i][nr_cols-1] = 
      ass_constraints->p[i][ass_constraints->nbcolumns-1];
    newm->p[i][eps_col] = 
      ass_constraints->p[i][eps_col];
  }

  newm->nonstrict = !newm->nplus1_is_strict(eps_col);
  assignment result_ass(nr_old_vars_used, oldvarindices, newm);
  result_ass.nr_new_vars = nr_new_vars_used;
  result_ass.new_indices = newvarindices;
  return(result_ass);
  // memory lost in oldvarpos
}



inline
assignment_info::assignment_info(assignment *ass) {
  asst = ass;
  next = NULL;
}


assignment_info *
assignment_info::insert(assignment_info *aip) {
  // insert into head of singly-linked list 
  aip->next = this;
  return(aip);
}


/* ===============  locations =============== */



location::location() {
  char *defaultname = "unknown";
  name = (char *) malloc( (strlen(defaultname)+1) * sizeof(char));
  strcpy(name, defaultname);
  copy_of_name = (char * ) NULL;
  nr_slopes = 0;
  post_slopes_matrix = NULL;
  pre_slopes_matrix = NULL;
  post_slopes = NULL;
  pre_slopes = NULL;
  nr_post_trans = 0;
  nr_pre_trans = 0;
  asap = false;
  post_trans = NULL;
  pre_trans = NULL;
  invariant = NULL;
  autp = (automaton *) NULL;
}

location::location(char *locn_name) {
  name = (char *) malloc( (strlen(locn_name)+1) * sizeof(char));
  strcpy(name,locn_name);
  copy_of_name = (char *) NULL;
  nr_slopes = 0;
  post_slopes_matrix = NULL;
  pre_slopes_matrix = NULL;
  post_slopes = NULL;
  pre_slopes = NULL;
  nr_post_trans = 0;
  nr_pre_trans = 0;
  asap = false;
  post_trans = NULL;
  pre_trans = NULL;
  invariant = NULL;
  autp = (automaton *) NULL;
}



location::location(const location & locn) {
  name = str_preserve(locn.name);
  copy_of_name = str_preserve(locn.copy_of_name);
  nr_slopes = locn.nr_slopes;
  post_slopes = locn.post_slopes;
  pre_slopes = locn.pre_slopes;
  post_slopes_matrix = locn.post_slopes_matrix;
  pre_slopes_matrix = locn.pre_slopes_matrix;
  nr_post_trans = locn.nr_post_trans;
  nr_pre_trans = locn.nr_pre_trans;
  post_trans = locn.post_trans;
  pre_trans = locn.pre_trans;
  asap = locn.asap;
  invariant = locn.invariant;
  autp = locn.autp;
}


void
location::set_pre_vectors() {
//
// assume have all the post slope vectors already there
// and that pre_slopes are non-existent as of yet.
//
// just make the pre-slope vectors, and link properly
//  assuming that all slopes are unique for now
// and set pre_slopes to point to array of ptrs to slopes

  int i,j;
  vector news;
  if (nr_slopes > 0) {
    int s = (*(post_slopes[0])).size;
    if (pre_slopes != NULL) {
//      cout << "Not assigning pre-slopes to location " << locn_id << 
//	       ": has slopes already\n";
    }
    else {
      pre_slopes = (slope_vector **) malloc (sizeof(slope_vector *) * s);
      for (i=0;i<nr_slopes; i++) {
	vector v = *(post_slopes[i]);
	news = vector(s);
	news.p[0] = v.p[0];  // should be a 1 for a ray!
	// negate slopes values, the last should be 0 already
	for (j=1;j<s;j++) {
	  news.p[j] = -v.p[j];
	}
	the_slopes = the_slopes->insert(new slope_vector_info(news));
	pre_slopes[i] = &(the_slopes->v);
      }
    }
  }
}

// set the asap field for this location
// use the post-trans info
void
location::set_asap_from_trans() {
  int i;
  asap = false;
  for (i=0;i<nr_post_trans;i++) {
    if (post_trans[i].is_asap_trans) {
      asap = true;
    }
  }
}



slope_vector **
location::get_slope_vectors_for(enum directions dir) {
  if (dir == PRE) {
    return (pre_slopes);
  }
  else {
    return( post_slopes);
  }
}

// modified 8/95 to avoid printing "Location: " is no name
void
location::name_print() {
  if ((name != NULL) && ( strlen(name) != 0)) 
    cout << "Location: " << name << "\n";
}

void
location::print() {
  int i;
  cout << "Location (" << locn_id << ") : " << name << "\n";

cout << "This = " << this << "\n";

  cout << "ASAP: " << Bool_output(asap) << "\n";
  cout << "Post slopes \n";
  cout << "Constraints: ";
  post_slopes_matrix->constraints()->print();

cout << "Eps pos =  " << post_slopes_matrix->eps_pos << "\n";

  if (pre_slopes_matrix != NULL) {
      cout << "Pre slopes\n";
      cout << "Constraints: ";
	// actually not a matrix but an opoly
      pre_slopes_matrix->constraints()->print();
    }
  else {
    cout << "No pre-slopes given\n";
  }

  if (invariant != NULL) {
    cout << "Invariant matrix\n";
    invariant->print();
  }
  else {
    cout << "No invariant matrix declared\n";
  }
  cout << "Number of post-trans = " << nr_post_trans << "\n";
  for (i=0;i<nr_post_trans;i++) {
    post_trans[i].print();
    put_line();
  }
  cout << "Number of pre-trans = " << nr_pre_trans << "\n";
  for (i=0;i<nr_pre_trans;i++) {
    pre_trans[i].print();
    put_line();
  }
}


void
location::free() {
  int i;
  global_free(name);
// using matrices now, not ptrs to slopes
//  global_free(post_slopes);
//  global_free(pre_slopes);
  if (post_slopes_matrix != NULL) {
    post_slopes_matrix->free();
    post_slopes_matrix = NULL;
  }
  if (pre_slopes_matrix != NULL) {
    pre_slopes_matrix->free();
    pre_slopes_matrix = NULL;
  }

  for (i=0;i<nr_post_trans;i++) {
    post_trans[i].free();
  }
  global_free(post_trans);
  for (i=0;i<nr_pre_trans;i++) {
    pre_trans[i].free();
  }
  if (pre_trans != NULL) {
    global_free(pre_trans);
  }
}


void
location::copy_loc_name() {
  global_free(copy_of_name);
  copy_of_name = str_preserve(name);
}


void
location::restore_loc_name() {
  global_free(name);
  name = copy_of_name;
  copy_of_name = NULL;
}



/* ===============  slope vectors =============== */


slope_vector_info::slope_vector_info(vector & vec) {
 v = vec;
 next = NULL;
}


slope_vector_info::slope_vector_info(){
/* get the default vector */
  v = vector(var_count+2);
  next = NULL;
}



slope_vector_info  *
slope_vector_info::insert(slope_vector_info * svip) {
  svip->next = this;
  return (svip);
}


void
slope_vector_info::print(){
/* get the default vector */
   v.print();
  if (next != NULL) {
    next->print();
  }
}


/* ============== component info =========== */

component_info::component_info() {
  automaton_name = "";
  nr_locations = 0;
  // is this NULL ptr what we want?
  locn_names = (char **) NULL;
}


// constructor to get info from the aut. make new copies of all strings
component_info::component_info(automaton *autp) {
  int i;
  automaton_name = str_preserve(autp->name);
  nr_locations = autp->nr_locations;
  // is this NULL ptr what we want?
  locn_names = (char **) malloc (nr_locations * sizeof (char *));
  for (i=0;i<nr_locations;i++) {
    locn_names[i] = str_preserve(autp->locations[i].name);
  }
}

void
component_info::print() {
   int i;
   cout << " Component info\n ==========\n\n";
   cout << "Automaton name: " << automaton_name << "\n";
   cout << "Nr locns: " << nr_locations << "\n";
   for (i=0;i<nr_locations;i++){
     cout << "\t" << locn_names[i] << "\n";
   }
   cout << "End component info ---------------\n";
 }


void
component_info::free() {
  int i;
  global_free(automaton_name);
  for (i=0;i<nr_locations;i++) {
    global_free(locn_names[i]);
  }
}

/* ===============  automaton =============== */


// automaton are initialized with an matrix for TRUE, and default number 
// of variables is taken from

automaton::automaton() {
  char *nom = "unnamed";
  nr_locations = 0;
  locations = NULL;
  name = (char *) malloc ((strlen(nom) + 1) * sizeof(char));
  strcpy(name, nom);

  // get TRUE matrix with var_count+2 cols
  matrix_hd = new matrix_info(var_count+2);
  nr_sync_labels = 0;
  init_location_p = NULL;
  init_region_p = NULL;
  sync_label_names = NULL;
  slope_lcm = 0;
  nr_slopes = 0;
  slopes = NULL;
  pre_is_valid = false;
  nr_components = 0;
  comp_inf_p = NULL;
  is_orig_matrix = NULL;
}


automaton::automaton(int s) {
// beware, locations are uninitialized
// only the space is allocated
  char *nom = "unnamed";
  nr_locations = 0;
  name = (char *) malloc ((strlen(nom) + 1) * sizeof(char));
  strcpy(name, nom);
 matrix_hd = new matrix_info(var_count+2);
  nr_sync_labels = 0;
  sync_label_names = NULL;
  init_location_p = NULL;
  init_region_p = NULL;
  locations =  (location *) malloc (s * sizeof(location));
  slope_lcm = 0;
  nr_slopes = 0;
  slopes = NULL;
  pre_is_valid = false;
  nr_components = 0;
  comp_inf_p = NULL;
  is_orig_matrix = NULL;
}  


automaton::automaton(char *nom, int s) {
// beware, locations are uninitialized
  nr_locations = 0;
  name = (char *) malloc ((strlen(nom) + 1) * sizeof(char));
  strcpy(name,nom);
  matrix_hd = new matrix_info(var_count+2);
  nr_sync_labels = 0;
  init_location_p = NULL;
  init_region_p = NULL;
  sync_label_names = NULL;
  locations =  (location *) malloc (s * sizeof(location));
  slope_lcm = 0;
  nr_slopes = 0;
  slopes = NULL;
  pre_is_valid = false;
  nr_components = 0;
  comp_inf_p = NULL;
  is_orig_matrix = NULL;
}  





automaton *
automaton::copy_of() {
 // get a ptr to a new automaton in free space.
  // however, shares same data in free space
  automaton * ap = new automaton();
  ap->nr_locations = nr_locations;
  ap->name = (char *) malloc ((strlen(name) + 1) * sizeof(char));
  strcpy(ap->name,name);
  // matrix_hd field not being used now.
  ap->matrix_hd = matrix_hd;
  ap->nr_sync_labels = nr_sync_labels;
  ap->init_location_p = init_location_p;
  ap->init_region_p = init_region_p;
  ap->sync_label_names = sync_label_names;
  ap->locations =  locations;
  ap->nr_slopes = nr_slopes;
  ap->slopes = slopes;
  ap->matrix_store = matrix_store;
  ap->nr_matrices_in_store = nr_matrices_in_store;
  ap->is_orig_matrix = is_orig_matrix;
  ap->id_slope_index = id_slope_index;
  ap->id_matrix_index = id_matrix_index;
  ap->slope_lcm = slope_lcm;
  ap->pre_is_valid = pre_is_valid;
  ap->nr_components = nr_components;
  ap->comp_inf_p = comp_inf_p;
  return(ap);  
}  


// shouldn't need to do this, since it is in the constructor
void
automaton::set_matrix_head() {
  matrix_hd = new matrix_info(matrix(nr_locations));
}

void
automaton::set_pre_vectors() {
  int i;
  for (i=0;i<nr_locations; i++) {
    locations[i].set_pre_vectors();
  }
}


matrix_info *
automaton::store_matrix_from_assignment(assignment inv_ass) {
  
  matrix_hd =
    matrix_hd->insert(new matrix_info(*(inv_ass.ass_constraints)));
  return ( matrix_hd );    
}

// free space for ids_for_slopes/matrices
// only call after doing the setting of global slopes and matrices lists
// and used the info for locations, transitions, etc.
void
automaton::free_id_vectors() {
//   global_free(id_slope_index);
  global_free(id_matrix_index);
}

// check the syntax/structure of the automaton
// 1) check that any events in set of synchronization labels appears at least
//    once in a transition
// 2) locations have incoming transitions
// 3) locations have outgoing transitions
// later add ...
// 4) initial conditions and guards on trans intersect invariants
// 5) if slope given for a var, then given for all locations?
//
// return degree of severity.  
//    2) and 3) are ignored unless both occur for 
//           a single given location.
//    1) is considered bad. return 1 for each such sync label
//  
//
// assumes the pre info for the automaton is not available
int
automaton::check_automaton() {
  int i,j;
  int synindex;
  Bool *label_used = (Bool *) malloc (sizeof(Bool) * nr_sync_labels);
  Bool *incoming = (Bool *) malloc (sizeof(Bool) * nr_locations);
  int severity = 0;  // default error level
  int beware;

  fprintf(stderr,"Checking automaton %s\n", name); 
  // initialize Bool arrays of check values
  for (i=0;i<nr_sync_labels;i++) {
    label_used[i] = false;
  }
  for (i=0;i<nr_locations;i++){
    incoming[i] = false;
  }

  // search over transitions
  for (i=0;i<nr_locations;i++) {
    for (j=0;j<locations[i].nr_post_trans;j++) {
      if (locations[i].post_trans[j].sync_label != (char *) NULL) {
	synindex = 
	  string_pos(locations[i].post_trans[j].sync_label,
		     sync_label_names,nr_sync_labels);
	if (synindex == -1) {
	  fprintf(stderr,"Sync label  %s not declared: aborting\n",
		          locations[i].post_trans[j].sync_label);
          exit(1);
	}
	else {
	  label_used[synindex] = true;
	}
      } // end if for sync label
      incoming[locations[i].post_trans[j].dest->locn_id] = true;
    }
  }

  // output any warnings
  for (i=0;i<nr_sync_labels;i++) {
    if (!label_used[i]) {
      fprintf(stderr,"\n\n WARNING: Sync. label %s declared but not used ",
	      sync_label_names[i]);
       fprintf(stderr,"in automaton %s\n\n", 
	       name);
      severity++;
    }
  }
  for (i=0;i<nr_locations;i++) {
    beware = 0;
    if (locations[i].nr_post_trans ==0) {
       beware += 1;
     }
    if (!incoming[locations[i].locn_id]) {
      beware +=2;
    }
    switch (beware) {
    case 0: 
       break;
    case 1: 
       cerr << " WARNING: locn " << locations[i].name << " of automaton "
	 << name << " has no outgoing transitions\n";
       break;
    case 2:
       cerr << " WARNING: locn " << locations[i].name << " of automaton "
	  << name << " has no incoming transitions\n";
       break;
    case 3:
       cerr << " WARNING: locn " << locations[i].name << " of automaton "
	  << name << " has neither incoming nor outgoing transitions\n";
       severity++;
       break;
    default: break;
    }
  }
  return(severity);
}


void
 automaton::add_inverse(transition &tr) {
//
// add the inverse transition.
//
// need to a) compute inverse transition (guard is TRUE). 
// b) allocate space for the assignment constraint matrix
//
// d) allocate space to place actual inverse transition
// e) set pointer in succ. location
//
// store inverse transitions with absolute (FW) references for src, dest.

  transition *inv_tr;
  matrix_info *mp;

  // 1) get the inverse assignment
  // ugly: how do I do this cleanly?
  // the assignment needs to store a constraint matrix in matrix list
  // which belongs to the automaton.  it cannot refer to it.
  // for now, make_inverse_ass() computes and stores matrix somewhere, 
  // then add it to list in 2nd step, tell assignment in 3rd step
  assignment *inv_ass = new assignment();
  *inv_ass  = tr.ass.make_inverse_ass();

  mp = store_matrix_from_assignment(*inv_ass);
  inv_ass->set_matrix_p(&(mp->m));

// cout << "Got this inverse assignment\n" << flush; 
// inv_ass->print();
// cout << flush;

  // 2) make the inverse transition
  inv_tr = new transition(tr.src, tr.dest, tr.sync_label, 
		       tr.enabling_constr, tr.is_asap_trans, inv_ass);

  // 3) place the inverse transition in next slot
  locations[tr.dest->locn_id].pre_trans[
	locations[tr.dest->locn_id].nr_pre_trans] = 
	  *inv_tr;
  locations[tr.dest->locn_id].nr_pre_trans++;
}


void
automaton::set_location_asap() {
  int i;
  for (i=0;i<nr_locations;i++) {
    locations[i].set_asap_from_trans();
  }
}


void
automaton::print() {
  int i;
  cout << "===============================================================\n";
  cout << "Automaton: " << name << "\n";
  cout << "Nr sync labels = " << nr_sync_labels << "\nSync labels are:\n";
  for (i=0;i<nr_sync_labels;i++) {
    cout << "\t" << sync_label_names[i] << "\n";
  }
  cout << "Nr locations = " << nr_locations << "\n";
  cout << "Init location is " << 
    ((init_location_p == NULL) ? "" : init_location_p->name);
  cout  << "\n";
  cout << "Init constraints are ";
  if (init_region_p != NULL)  init_region_p->print() ; 
  cout << "\n" << flush;
  for (i=0;i<nr_locations;i++) {
    locations[i].print();
  }
//  cout << "matrices on matrix_hd (supposed global matrix list) is \n";
//  matrix_hd->print();
  cout << "Nr of component automata: " << nr_components << "\n";
  for (i=0;i<nr_components;i++) {
    comp_inf_p[i].print();
  }
}

// output #locns, #pretrans, #posttrans
// # matrices in store, #vectors
void
automaton::print_size() {
  int i;
  int num_post = 0;   // total number of post transitions
  int num_pre = 0;    // total number of pre transitions
  int nr_rows = 0;
  int nr_orig = 0;

  cout << "\n";
  cout << "Size of Automaton: " << name << "\n";
  cout << "Nr locations = " << nr_locations << "\n";
  cout << "Nr matrices stored = " << nr_matrices_in_store << "\n";
  cout << "Nr vectors stored = " << nr_slopes << "\n";
  cout << "Nr global variables = " << var_count << "\n";
  cout << flush;

  // count # post_trans & matrix sizes
  for (i=0;i<nr_locations;i++) {
    num_post += locations[i].nr_post_trans;
    num_pre += locations[i].nr_pre_trans;
    nr_rows += matrix_store[i].nbrows;
    if (is_orig_matrix[i]) {
      nr_orig++;
    }
  }
  
  cout << "Nr post trans = " << num_post << "\n";
  cout << "Nr pre trans = " << num_pre << "\n";
  cout << "Nr orig matrices = " << nr_orig << "\n";
  cout << "Total number rows in invariant matrices = " << nr_rows << "\n";
  if (nr_locations >0) {
    cout << "Average number rows in invariant matrices = " << 
	nr_rows/nr_locations << "\n";
  }
  cout << "\n\n" << flush;

}

void
automaton::set_pre_trans() {
//
// assume we have all the post info already
//
// do two passes. the first time to count the 
// transitions into each location, the second to actually assign the 
// information.  in between the 2 passes, set up the data structures for
// the pre_trans().  this saves wasting extra pointers, etc, and unecessary 
// copying later.
//
// could alternatively compute the pre transitions as a linked list
// and then copy over.
//
  int i,j;

  // pass 1. count number pre trans per location
  for (i=0;i<nr_locations; i++) {
    locations[i].nr_pre_trans = 0;
  }
  
  for (i=0;i<nr_locations;i++) {
    for (j=0;j<locations[i].nr_post_trans;j++) {
      locations[locations[i].post_trans[j].dest->locn_id].nr_pre_trans++;
    }
  }

  // assign memory required, we use ptr to array of ptrs to transitions
  for (i=0;i<nr_locations;i++) {
    locations[i].pre_trans = 
      (transition *) malloc ((locations[i].nr_pre_trans) * sizeof(transition));
   locations[i].nr_pre_trans = 0;  // none stored there yet.
  }

  // pass 2. create the new transitions 
  
  for (i=0;i<nr_locations;i++) {
    for (j=0;j<locations[i].nr_post_trans;j++) {
       add_inverse(locations[i].post_trans[j]);
    }
  }
}


void
automaton::set_pre_slope_matrices() {
  int i,j,k;
  for (i=0;i<nr_locations;i++) {
    matrix *fr  = 
	new matrix(*(locations[i].post_slopes_matrix->frame()->copy()));
    for (j=0;j<fr->nbrows-1;j++) {
      // negate all values except first and last columns
      for (k=1;k<fr->nbcolumns-1;k++) {
	fr->p[j][k] = -fr->p[j][k];
      }
    }
    opoly * pre_s_m = new opoly();
    pre_s_m->nplus1_form = new lib_polyhedron(*fr,false);
    pre_s_m->eps_pos = fr->nbcolumns-2;
    pre_s_m->strict = false;
    locations[i].pre_slopes_matrix = pre_s_m;

    global_free(fr);
  }
}

void
automaton::set_pre() {
// this will later be made the only accessible routine wrt making pre()
// both subroutines, and hence also this, assume all post info already present
  if (!pre_is_valid) {
//     cerr << "Setting pre info\n" << flush;
    set_pre_slope_matrices();
    set_pre_trans();
    pre_is_valid = true;
  }
}

// freeing automaton components


void
automaton::free_component_info() {
  int i;
  for (i=0;i<nr_components;i++) {
    comp_inf_p[i].free();
  }
  global_free(comp_inf_p);
}

void
automaton::free_matrix_store() {
  int i;

  for (i=0;i<nr_locations;i++) {
//   changed 12/15/95
//    if ((is_orig_matrix == NULL) || is_orig_matrix[i]) {
    if ((is_orig_matrix != NULL) && is_orig_matrix[i]) {
//    matrix_store[i].print();
      matrix_store[i].free();
    }
  }
  for (i=nr_locations;i<nr_matrices_in_store;i++) {
    matrix_store[i].free();
  }
  global_free(matrix_store);
  if (is_orig_matrix != NULL) {
    global_free(is_orig_matrix);
    // added 12/15/95
    is_orig_matrix = NULL;
  }
}

void
automaton::free_slopes() {
  if (slopes != NULL) {
    cout << "freeing slopes value = " << slopes << "\n" << flush;
    global_free(slopes);
  }
}

// free locations, after freeing information in their transitions
void
automaton::free_locations() {
  int i;
  for (i=0;i<nr_locations;i++) {
    locations[i].free();
  }
  global_free(locations);
}

void
automaton::free() {
  free_component_info();
//  free_matrix_store();
  free_slopes();
  free_locations();
}

// copy the names of the locations
// need for the nasty  simplification routines that eliminate locs.
void 
automaton::copy_location_names() {
  // use component info to figure out the name of each location.
  // this is very dependent on how the product is formed.
  // it would be better to have the component location info stored correctly.
  // per locn.
  int i;
  for (i=0;i<nr_locations;i++) {
    locations[i].copy_loc_name();
  }
}

// restore the names of the locations, using the copy_of_name field
// needed for after the routines that change the names of the locations
// for simplification.  nasty.  also, sets copy_of_name field to NULL
void 
automaton::restore_location_names() {
  // use component info to figure out the name of each location.
  // this is very dependent on how the product is formed.
  // it would be better to have the component location info stored correctly.
  // per locn.
  int i;
  for (i=0;i<nr_locations;i++) {
    locations[i].restore_loc_name();
  }
}


//  *************  lists of automata ********************

automaton_info::automaton_info() {
  a = new automaton();  
  next = NULL;
}


automaton_info::automaton_info(automaton aut) {
  a = aut.copy_of();   // copy ptrs only, not new data
  next = NULL;         // for list of lists, i.e. only concerned with 1 list
                       // pted to by a.
}


automaton_info *
automaton_info::insert(automaton_info *aip) {
  // insert into head of NON-EMPTY singly-linked list 
  aip->next = this;
  return(aip);
}

void
automaton_info::print() {
  if (a!= NULL) {
    a->print();
  }
  if (next != NULL) {
    next->print();
  }
}


// compose the list of automata (assumed nonempty)
// return singleton list with 1 automaton for the product
automaton_info *
automaton_info::compose() {
// repeatedly compose with next automaton, if it exists, and 
// 
  automaton_info *aip = this;
  automaton *autp;
  int i;

  cerr << "Composing automata " << flush;

  if (aip->next == NULL) {
    // one automaton only

    a->add_comp_info(0,a);
    for (i=0;i<a->nr_locations;i++) {
      a->locations[i].autp = a;
    }
    cerr << "\n" << flush;
    return(this);
  }
  else {
    autp = a->copy_of();
    while (aip->next != NULL) {
      automaton *oldautp = autp;
      aip = aip->next;
      autp = (autp->compose(*(aip->a))).copy_of();
      // 1/97 bug fix hack, since reach_cpomopse gives incorrect data for autp
      // now since it is a ptr.
      for (i=0;i<autp->nr_locations;i++) {
	autp->locations[i].autp = autp;
      }
      cerr << ".";
      cerr << "." << flush;
      oldautp->free();
      delete oldautp;
    }
    cerr << "\n" << flush;
    automaton_info * result =  new automaton_info(*autp);
    // 1/97 bug fix hack, since reach_comopse gives 
    //       incorrect data for autp field of locations.
    // now since it is a ptr.
    for (i=0;i<autp->nr_locations;i++) {
      result->a->locations[i].autp = result->a;
    }
    return(result);
  }
}



// compose the list of automata (assumed nonempty)
// return singleton list with 1 automaton for the product
automaton_info *
automaton_info::reach_compose() {
// repeatedly compose with next automaton, if it exists, and 
// 
  automaton_info *aip = this;
  automaton *autp;
  int i;

  cerr << "Composing automata " << flush;

  if (aip->next == NULL) {
    // one automaton only
    a->add_comp_info(0,a);
    for (i=0;i<a->nr_locations;i++) {
      a->locations[i].autp = a;
    }
    cerr << "\n" << flush;
    return(this);
  }
  else {
    autp = a->copy_of();
    while (aip->next != NULL) {
      automaton *oldautp = autp;
      aip = aip->next;
      autp = (autp->reach_compose(*(aip->a))).copy_of();
      // 1/97 bug fix hack, since reach_compose gives 
      //   incorrect data for autp now since it is a ptr.
      for (i=0;i<autp->nr_locations;i++) {
	autp->locations[i].autp = autp;
      }
      cerr << "*" << flush;
      oldautp->free();
      delete oldautp;
    }
    cerr << "\n" << flush;
    automaton_info * result =  new automaton_info(*autp);
    // 1/97 bug fix hack, since reach_compose gives incorrect data for autp
    // now since it is a ptr.
    for (i=0;i<autp->nr_locations;i++) {
      result->a->locations[i].autp = result->a;
    }
    return(result);
  }
}


Bool
test_ok_to_generate(automaton *a) {
  int i;
  Bool OK = true;
  for (i=0;i<a->nr_locations;i++) {
    if (!ok_to_generate(a->locations[i].post_slopes_matrix->nplus1_form)) {
      cout << "Location " << a->locations[i].name << 
	" has illegal rate conditions -- see the user guide\n";
      OK = false;
    }
  }
  return (OK);
}

