
#ifndef __CONVERT_H
#define __CONVERT_H

#include "poly.H"          
#include "include.h"
#include "location.h"
#include "output.h"


#include "extrapoly.H"    
#include "automaton.H"
#include "opoly.h"


//  brief documentation on structural changes in moving automaton from
// C to C++.
//
// mainly menial, except in assignments, where we move from multiple
// assignments into intervals, to a single matrix for resulting constraints. 
//
// C internal structure
//
// a linked list of automata.  items are generally numbered globally across
//    all automata. references to lists is via index numbers
// each has linked lists for vectors, matrices, and locations
//   listed ordered in strictly decreasing order (?, at least mainly).
//   vectors: used for assignments and slopes
//   matrices: each matrix has a linked list of vectors. in reverse order!
//   locations: locn_id
//              matrix_idx (invariant)
//              slope_vector start_idx
//              slope_vector end_idx
//              linked list of transitions
//       transitions: trans_id
//                    guard idx
//                    linked list of assignments
//               assignments: var_id
//                            lower bound vect idx
//                            upper bound vect idx
//
//
// assignment vectors: 
//              uses halbwachs notation, except denom, and nom.
//              
//
// C++ structure
//
//
// array of ptrs to single automata
//
// for a single automaton
//
// nr_slope_vecs
// slopes: can be an array of slope vectors.
// automaton:
//      nr_locations:
//      matrix_hd: (type matrix_info) linked list of nodes with ptrs to 
//                                                       matrices
//      locations: ptr to (array of) locations. 
//          each location:
//               nr_slopes.
//               pre_slopes,post_slopes: ptr to array of ptrs to slope vectors.
//               invariant: ptr to matrix for invariant
//               post_trans, pre_trans:
//                  each a ptr to array of transitions.
//                  each transitions:
//                        src, dest: ptr to location (viewed in abs fw dirn)
//                        sync_label: char *
//                        enabling_constr: ptr to matrix
//                        ass: assignment
//                              nr_vars:
//                              indices: ptr to array of indices of 
//                                           assigned vars.
//                              ass_constraints: ptr to matrix for constraints.
//                                           where we extend matrix with cols
//                                           for primed versions of ass. vars.
//
//
// assume the C to C++ conversion is through an automaton called C_automaton
//
void Cpp_matrix(matrix *mcpp, struct _p_matrix *m);
void Cpp_matrix_nplus1(matrix *mcpp_nplus1, struct _p_matrix *m);
void make_var_names();  // do this in main.C, to capture the var eps-slack too
#endif
