#include "poly.H"
#include "extrapoly.H"
#include "region.H"
#include "opoly.h"

// #include "include.h"
// #include "convert.H"
// #include "automaton.H"


// ===============================================
//
// Section 1. Later to be moved to extra_opoly.C
//
// provides implementation for
//
// opoly::complement       complement for a single element

// ===============================================
//
// Section 2. Later to be moved to region.C
//
// provides implementation for
//
// region::intersect_list
// region::list_intersect_list
// region::complement       complement of a single element (?)
// region::complement_list  complement of a whole list
// region::minus            object - argument  (entire lists)
//
// region_array::ra_complement
// region_array::ra_minus
//

// ===================================================================
//
// Section 1. Later to be moved to extra_opoly.C
//




// yield list for intersection of object with argument's list
// list is new data, independent of object and its argument
region *
region::intersect_list(region *int_list) {
  region *result = NULL;
  region* single_int;
  region *rp;
  
  rp=int_list;
  while (rp) {
    if (!rp->is_empty()) {
      single_int = rp->intersect(this);
      region_or(&result, single_int);
      if (single_int) {
	single_int->free();
      }
    }
    rp = rp->next;
  }
  return result;
}



// yield list for intersection of object's list with argument's list
// list is new data, independent of object and its argument
region *
region::list_intersect_list(region *int_list) {
  region *result = NULL;
  region *single_int_list;
  region *rp;
  
  rp=this;
  while (rp) {
    if (!rp->is_empty()) {
      single_int_list = rp->intersect_list(int_list);
      region_or(&result, single_int_list);
      if (single_int_list) {
	single_int_list->free_list();
      }
    }
    rp = rp->next;
  }
  return result;
}




// yield list of polys for the complement of the single object poly (not list)
// assumes poly is nonempty
region*
region::complement() {
  int i,j,k;
  int dimension = 0;
  region* result = NULL;

  if (is_empty()) {
    if (polyp) {
      dimension = polyp->dim();
    }
    // return the poly for true
    matrix* m = new matrix(2,dimension+2);
    for (i=0;i<dimension+2;i++) {
      m->p[0][i] = m->p[1][i] = 0;
    }
    m->p[0][0] = m->p[0][dimension+1] = 0;
    m->p[1][0] = m->p[0][dimension+1] = 1;
    opoly* op = new opoly(*m);
    return new region(locnp,op);
  }
  dimension = polyp->dim();
  matrix *n1cons = polyp->nplus1_form->constraints();
  int nrows = n1cons->nbrows;
  int ncols = n1cons->nbcolumns;
  int dim = ncols-3;

  for (i=0;i<nrows-1;i++) {
    for (j=1;j<ncols-2;j++) {
      if (n1cons->p[i][j]) {
	// this row depends on some real (non-epsilon) variable
	// complement it
	if (n1cons->p[i][0] == EQUALITY_CONSTANT) {
	  // complement has two parts
	  if (n1cons->p[i][ncols-2]) {
	    cout << "Error?  Equality row with epsilon and x_i\n";
	    exit(1);
	  }
	  // m1 is the a1*x1+...+an*xn+a(n+1)>0 row
	  matrix* m1 = new matrix(2,ncols-1);
	  m1->p[0][0] = 2;
	  for (k=1;k<=dim;k++) {
	    m1->p[0][k] = n1cons->p[i][k];
	  }
	  m1->p[0][dim+1] = n1cons->p[i][dim+2];
	  m1->set_last_row();
	  
	  // m2 is the a1*x1+...+an*xn+a(n+1)<0 row
	  matrix* m2 = new matrix(2,ncols-1);
	  m2->p[0][0] = 2;
	  for (k=1;k<=dim;k++) {
	    m2->p[0][k] = - n1cons->p[i][k];
	  }
	  m2->p[0][dim+1] = - n1cons->p[i][dim+2];
	  m2->set_last_row();
	
	  region* reg1 = new region(locnp,new opoly(*m1));
	  region* reg2 = new region(locnp,new opoly(*m2));
	  region_or(&result,reg1);
	  region_or(&result,reg2);
	  reg1->free(); // global_free(reg1);
	  reg2->free(); // global_free(reg2);
	}
	else {
	  // row denotes an inequality constraint
	  matrix *m1 = new matrix(2,ncols-1);
	  if (n1cons->p[i][ncols-2]) {
	    // row denotes strict inequality
	    // hence complement is nonstrict
	    m1->p[0][0] = NONSTRICT_CONSTANT;
	  }
	  else {
	    // row denotes nonstrict inequality
	    // hence complement is strict
	    m1->p[0][0] = STRICT_CONSTANT;
	  }
	  for (k=1;k<=dim;k++) {
	    m1->p[0][k] = - n1cons->p[i][k];
	  }
	  m1->p[0][dim+1] = - n1cons->p[i][dim+2];
	  m1->set_last_row();
	  
	  region* reg1 = new region(locnp,new opoly(*m1));
	  region_or(&result,reg1);
	  reg1->free(); // global_free(reg1);
	}
      break;  // out of  for (j...) loop for testing across row for nonzero elt
      }
    }
  }
  return(result);
}


// ===================================================================
//
// Section 1. Later to be moved to region.C
//




// yield the complement of a whole list of polys pted to by object
region*
region::complement_list() {

  region *temp;
  region *result = true_region(locnp, var_count+2);
  region *comp_of_single = NULL;

  region *rp = this;
  while (rp) {
    if (!rp->is_empty()) {
      comp_of_single = rp->complement();
      temp=result;
      result = result->list_intersect_list(comp_of_single);
      if (temp) {
	temp->free_list();
      }
      if (comp_of_single) {
	comp_of_single->free();
      }
    }
    rp = rp->next;
  }
  return(result);
}



region*
region::region_and(region* rap) {

  return(list_intersect_list(rap));

}

// yield list of regions for object's list "minus" (set diff.) argument's list
region*
region::minus(region *rp) {
  region * result;
  if (rp == NULL) {
    result = fresh_copy_list();
  }
  else {
    region* arg_comp_list = rp->complement_list();
    // cout << "here's arg com list" << endl << flush;
    //    if (arg_comp_list) 
    //      arg_comp_list->print();
    // cout << "end of arg comp list\n" << flush;
    result = region_and(arg_comp_list);
    if (arg_comp_list) {
      arg_comp_list->free_list();
    }
  }
  //  cout << "here's result of minus, with ptr " << result << "\n" << flush;
  // if (result)
  //   result->print();
  // cout << "end of res\n" << flush;
  return(result);
}



//  yield complement of object
region_array*
region_array::ra_complement() {
  int i;
  region_array* result = new region_array(size, autp);
  for (i=0;i<size;i++) {
    if (regions[i] && !regions[i]->is_empty_list()) {
      result->regions[i] = regions[i]->complement_list();
    }
    else {
      result->regions[i] = 
	true_region(&(autp->locations[i]), var_count+2);
    }
  }
  return(result);
}

// yield difference of object minus argument
region_array*
region_array::ra_minus(region_array* rap) {

  region_array* arg_compl = rap->ra_complement();
  region_array* result = ra_and(arg_compl);
  arg_compl->free();
  return(result);
}




region *
true_region(location * locp, int sz) {
  // the opoly operator automatically adds an extra column for the slack
  // variable eps
  matrix *m = new matrix(1,sz-1);
  m->set_last_row();
  polyhedron* pp = new opoly(*m);
  return(new region(locp,pp));
}


