#include "opoly.h"
#include "extrapoly.H"


int opoly::empty()
{
	int k = eps_pos;
	matrix* f = nplus1_form->frame();
        if ((k<=0) || k>= f->nbcolumns-1) {
	  cerr << "illegal value of eps_pos: " << k << "\n";
	  exit ( 1 );
	}
	strict = !constraints()->nplus1_def_nonstrict(k);

	if (!strict) {
	  nplus1_form->empty = check_empty(nplus1_form->frame()) ;
  	  return (nplus1_form->empty);
	}
	else 
	{ 
	  // may have strict matrix, without explicit eps constraints
	  add_eps_constraints();
	  f = nplus1_form->frame();
  	  for (int i = 0; i < f->nbrows; i++)
  	  {
	    if (( f->p[i][k] > 0) &&  (f->p[i][f->nbcolumns-1] > 0))
	      // nonempty
	      return 0;
	    else {
	      if (( f->p[i][k] < 0) &&  (f->p[i][f->nbcolumns-1] < 0)) {
		cerr << "fixes previously undetected emptiness\n";
		return 0;
	      }
	      else {
		if  (( f->p[i][k] > 0) &&  (f->p[i][f->nbcolumns-1] < 0)) {
		  cerr << "fixes previously incorrect emptiness evaluation\n";
		}
	      }
	    }
	  }
	  return 1;
	}

}



opoly *
opoly::existent(int vnr, int nr_extra_vars) {
   nplus1_form = nplus1_form->existent(vnr,nr_extra_vars);
// cout << "in opoly existent: result->nplus1_form() " << nplus1_form
// 	<< "\n" << flush;
  return this;
}



opoly *
opoly::existent(int nr_exist_vars, int *indices) {
   nplus1_form = nplus1_form->existent(nr_exist_vars, indices);
   return this;
}


// opoly *
// opoly::CopyColumn(int i2, int i3) {
//   
// }

opoly *
opoly::copy_extra_to_old(int nr_keep_cols, int nr_copy_cols, 
				 int *index_p) {
  nplus1_form = nplus1_form->copy_extra_to_old(nr_keep_cols, 
						  nr_copy_cols,
						  index_p);
  return this;
}
	


opoly *
opoly::vaporize_variables(int nr_v) {
// note: maintain same eps_pos
  return( new opoly(nplus1_form->vaporize_variables(nr_v),
		    POLYHEDRON_CONSTANT, (dim()-nr_v)-2-eps_pos) );
}

// 
// void
// opoly::ptprint() {
//   nplus1_form->ptprint();
// }

opoly *
opoly::hull(opoly *op) {
  // take the hull of the closures for now
  return(new opoly(closure(NPLUS1_FORM)->hull(op->closure(NPLUS1_FORM)),
		   POLYHEDRON_CONSTANT,0));
}

opoly *
opoly::add_row(matrix *mp, int rownr, Bool negated) {
// assume matrix in n+1 form
  return(new opoly(nplus1_form->add_row(mp, rownr,negated),
		   POLYHEDRON_CONSTANT,0));
}



Bool
operator == (opoly& op1, opoly& op2) {
  return( (op1 <= op2) && (op2 <=op1) );

}

opoly*
empty_polyhedron(int nrcols) {
  return(new opoly(empty_lib_polyhedron(nrcols),
		   POLYHEDRON_CONSTANT,0));
}

opoly * 
Resize_opoly(opoly *pp4, int newsize) {
  return(new 
	opoly(Resize(pp4->nplus1_form,newsize),
		   POLYHEDRON_CONSTANT,newsize-2-pp4->eps_pos));
}

// pretty print borrowed from extrapoly.C with modifications for
//   strict inequalities
void
opoly::ptprint() {
  if (empty()) {
    cout << "False\n";
  }
  else {
    nplus1_form->constraints()->ptprint(NPLUS1_FORM);
  }
}


matrix *
nplus1_rates_matrix(const matrix &m) {

   matrix  * np1 = new matrix(m.nbrows+1,m.nbcolumns+1);


/*
 *	np1->p[0] is the 0 = \dot{epsilon} row
 */


	for (int j = 0; j < m.nbcolumns+1; j++)
		np1->p[0][j] = 0;
	np1->p[0][m.nbcolumns-1] = 1;


/*
 *	for each remaining row, add epsilon term if strict
 */

	for (int i = 1; i < m.nbrows+1; i++)
	{
		{
			for (int j = 0; j < m.nbcolumns - 1; j++)
				np1->p[i][j] = m.p[i-1][j];
			np1->p[i][m.nbcolumns - 1] = 0;
			np1->p[i][m.nbcolumns] = m.p[i-1][m.nbcolumns-1];
		}
	}
	return np1;
}

// resize the matrix by adding a column for the eps_slack variable
// in index position eps_pos
matrix *
nplus1_plain_matrix(const matrix &m, int eps_pos) {

   matrix  * np1 = new matrix(m.nbrows,m.nbcolumns+1);
   np1->nonstrict = true;
/*
 *	for each row, add epsilon term if strict
 */

  if ((eps_pos<=0) || (eps_pos > m.nbcolumns-1)) {
	cerr << "Illegal call of nplus1_plain_matrix, with insert position " <<
	  eps_pos << "\n";
	exit ( 1 );
  }

  for (int i = 0; i < m.nbrows; i++)
    {
      int j;
      for (j = 1; j < eps_pos; j++)
	np1->p[i][j] = m.p[i][j];
      for (j = eps_pos+1; j < m.nbcolumns+1; j++) {
	np1->p[i][j] = m.p[i][j-1];
      switch (m.p[i][0]) {
	case EQUALITY_CONSTANT : 
	    np1->p[i][0] = np1->p[i][eps_pos] = 0;
	    break;
	case NONSTRICT_CONSTANT :
	    np1->p[i][0] =NONSTRICT_CONSTANT;
            np1->p[i][eps_pos] = 0;
	    break;
	case STRICT_CONSTANT:
	    np1->p[i][0] = NONSTRICT_CONSTANT;
	    np1->p[i][eps_pos] = -1;
	    np1->nonstrict = false;
	    break;
	  default: break;
	}
     }
  }
  return np1;
}



opoly *  
opoly::intersection(const matrix& constraints, Bool in_nplus1_form) 
{

  if (in_nplus1_form) {
    if (!nplus1_form->constraints()->nplus1_is_strict(eps_pos) && 
	!constraints.nonstrict) {
      add_eps_constraints();
    }
    nplus1_form->intersection(constraints);
    // may come back empty
    if (empty()) {
      int this_dim = dim();
      nplus1_form->free();
      delete nplus1_form;
      nplus1_form = empty_lib_polyhedron(this_dim);
      return this;
    }

   strict = nplus1_form->constraints()->nplus1_is_strict(eps_pos);
    nplus1_form->constraints()->nonstrict = !strict;
// cout << "simplifying\n";
//   print();
    if (!strict)
      simplify(); 
// cout << "gives\n";
// nplus1_form->print();
// cout << "printing poly\n";
//     print();
// cout << "done\n";
  }
  else {
    matrix* m_n1 = my_nplus1_matrix(constraints);
    nplus1_form->intersection(*m_n1);
    strict = nplus1_form->constraints()->nplus1_is_strict(eps_pos);
    nplus1_form->constraints()->nonstrict  = !strict;
     m_n1->free();
    delete m_n1;
  }
  return this;
}


// same as intersection, except simplify() commented out
opoly *  
opoly::intersection_no_simp(const matrix& constraints, Bool in_nplus1_form) 
{
  if (in_nplus1_form) {
    nplus1_form->intersection(constraints);
    strict = nplus1_form->constraints()->nplus1_is_strict(eps_pos);
// simplify();
  }
  else {
    matrix* m_n1 = my_nplus1_matrix(constraints);
    nplus1_form->intersection(*m_n1);
    strict = nplus1_form->constraints()->nplus1_is_strict(eps_pos);
    m_n1->free();
    delete m_n1;
  }
  return this;
}



// if nonstrict, simplify the nplus1_form by eliminating constraints 
// 0 <= eps_slack <= 1
void
opoly::simplify() {
//   return;
  strict = constraints()->nplus1_is_strict(eps_pos);
  constraints()->nonstrict = !strict;
  // HWT 6/25/98: only do so if (var_count>0), i.e. dimension > 3
  // or else get a problem with empty polys, where the
  // negativity contradictory rows are over the eps_slack variable
  // and these get eliminated, yuielding a true matrix from a false one!!
  if (!strict && constraints()->eps_present(eps_pos) && (nplus1_form->dim()>3)) {
    // get simplified form by eliminating info re eps_slack
    
    lib_polyhedron * simp_nplus1_form = closure(1);
//    cout << "within simpl: simplifying\n";
//    cout << "Eps pos = " <<   eps_pos << "\n";
//    constraints()->print();

    nplus1_form->free();
    delete nplus1_form;
    nplus1_form = simp_nplus1_form;
// cout << "within simp: nplus1_form = " << nplus1_form << "\n" << flush;
//   cout << "within simp:simplified = \n";
//    nplus1_form->print();
 
//    nplus1_form->constraints()->print();
//   cout << "within simp: done printing  simplified\n";
  }
}



int
opoly::dim() {
  if (nplus1_form) {
    return nplus1_form->dim();
  }
  else {
    cout << "Error: call on dim() on empty opoly\n" ;
    exit (1);
  }
}


void
opoly::fix_strict() {
  strict = constraints()->nplus1_is_strict(eps_pos);
}



void
opoly::minimize() {
  nplus1_form->minimize();
}

void
opoly::add_eps_constraints() {
// first do a check that they're not explicitly there already
  Bool eps_cons_there = false;
  matrix *cons = constraints();

  int i,j;
  for (i=0;i<cons->nbrows-1;i++) {
    if (!eps_cons_there && cons->p[i][eps_pos]) {
      if (cons->p[i][0] != 1) 
	goto iend;
      for (j=1;j<eps_pos;j++) 
	if (cons->p[i][j]!=0) 
	  goto iend;
      for (j=eps_pos+1;j<cons->nbcolumns-1;j++) 
	if (cons->p[i][j] != 0)
	  goto iend;
      if ((cons->p[i][eps_pos] == 1) && (cons->p[i][cons->nbcolumns-1]==0)) {
	eps_cons_there = true;
      }
      else {
	if ((cons->p[i][eps_pos] == -1) && (cons->p[i][cons->nbcolumns-1]==1)){
	  eps_cons_there = true;
	}
      }
    }
  iend: ;
  }
  if (!eps_cons_there) {
    matrix newm = matrix(cons->nbrows+2, cons->nbcolumns);
    for (i=0;i<cons->nbrows-1;i++) {
      for (j=0;j<cons->nbcolumns;j++) {
	newm.p[i][j] = cons->p[i][j];
      }
    }
    for (j=1;j<cons->nbcolumns-1;j++) {
      newm.p[cons->nbrows-1][j] = newm.p[cons->nbrows][j]=0;
    }
    newm.p[cons->nbrows-1][0] = newm.p[cons->nbrows][0] = 1;
    newm.p[cons->nbrows-1][eps_pos] = 1;
    newm.p[cons->nbrows-1][newm.nbcolumns-1] = 0;
    newm.p[cons->nbrows][eps_pos] = -1;
    newm.p[cons->nbrows][newm.nbcolumns-1] = 1;
    newm.set_last_row();

// cout << "Adding eps constraints to matrix in eps-pos = \n" << eps_pos << "\n";
// cons->print();

    nplus1_form->free();
    delete nplus1_form;
// cout << "yielding\n";
// newm.print();
    nplus1_form = new lib_polyhedron(newm, true);
  }
}



/*
// just add an extra row in the position 1 to the left of the constant
matrix *
ns_nplus1_matrix(const matrix &m) {
  int i;
  ns_matrix* nsp = new ns_matrix(m.nbrows,m.nbcolumns+1);
  for (i=0;i<m.nbrows;i++) {
    for (j=0;j<m.nbcolumns-1;j++) {
      nsp->p[i][j] = m->p[i][j];
    }
    nsp->p[i][m.nbcolumns-1] = 0;
    nsp->p[i][m.nbcolumns] = m->p[i][m.nbcolumns-1];
  }
  return nsp;
}
*/

// input matrix is in n form
Bool
is_strict_matrix(matrix m) {
  return (m.is_strict());
}


// compute whether input n matrix is strict, 
// if strict return strict form with
//    0<=eps<=1 constraints added
// else return extended form only
matrix *
my_nplus1_matrix(const matrix &m) {

  if (is_strict_matrix(m)) {
    return( nplus1_matrix(m) );
  }
  else {
    return( nplus1_plain_matrix(m,m.nbcolumns-1) );
  }
}


// use this copy function where the opoly is being used for slopes, i.e.
//   don't use the checks for emptiness, etc.
opoly*	
opoly::copy_for_slopes() {
  opoly * cp = new opoly();
  cp->nplus1_form = nplus1_form->copy();
  cp->eps_pos = eps_pos;
  cp->strict = strict;
  return cp;
}
