
%{
#include "include.h"
#include "parser.h"
#include "location.h"
#include "output.h"
#include "basic.h"
#include "predicate.h"
#include "spec.h"

extern FILE *yyin;

/* to overcome yacc errors on DECs for redefining YYSTYPE */
#define YYSTYPE HWT_YYSTYPE 

/* some versions of yacc use this as a bound on the size of the parse stack */
/* Sun versions seem to ignore it */
#define YYMAXDEPTH 2000 



%}        
 

%start main_program


%union	yystacktype {
  char		*string;
  int		integer;
  enum variable_type v_type;
}


%token ANALOG
%token ASAP
%token AUTOMATON
%token ASSIGNMENT
%token CLOCK
%token DISCRETE
%token DO
%token END
%token P_FALSE
%token GOTO
%token IN
%token INTEGRATOR
%token INT
%token LOC
%token NAME
%token PARAMETER
%token PRIM 

%token SYNC
%token SYNCLABS
%token P_TRUE
%token VARS
%token WAIT
%token WHEN
%token WHILE
%token INITIALLY
%token INFINITY
%token DIRECTION
%token P_PRE
%token P_POST
%token ELIM_ALL_VARS
%token ELIM_ALL_LOCNS
%token ELIM_VARS
%token ELIM_LOCNS
%token FINAL
%token EQUALS
%token VAR
%token REGION
%token FREE
%token PRINT
%token PRINTSIZE
%token ITERATE
%token FROM
%token USING
%token REACH
%token ENDREACH
%token FORWARD
%token BACKWARD
%token REGNAME
/* %token OUTSTRING */
%token PRINTSIMPLIFIED
%token PRINTSTRING
%token ALL
%token HIDE
%token ENDHIDE
%token OMIT
%token REG_OMIT
%token NON_PARAMETERS
%token LOCATIONS
%token IF
%token THEN
%token ELSE
%token ENDIF
%token EMPTY
%token HULL
%token STRING
%token ENDWHILE
%token P_DIFF
%token TRACE
%token TO
%token <integer> RELOP
%token <integer> WEAKLE

%type <integer> integer

%nonassoc WEAKLE
%nonassoc WEAKEQ
%nonassoc WEAKGE
%left P_OR
%left P_AND
%nonassoc RELOP, P_LT, P_GT
%right '~'   /* negation over regions == set complement */
%left '|'
%left '&'
%nonassoc P_WEAKDIFF
%right P_NOT   /* Boolean not */
%%
/* ordering of declaration of OR/AND gives AND higher priority */

main_program : { sync_number = 1; var_count = 0; } declarations
               { columns = var_number + 1; matrix_id = -1; some_slope = 1;
                 vector_id = loc_number = trans_num = 0; slope_init();
	       }
               program
               specification
             ; 
program :  AUTOMATON NAME {store_automaton_name((char *)yytext);}
           automaton 
          {
          if (SIMPLE_OUTPUT) {
            printf(" DONE --------------------\n"); 
            print_var_names1();
            print_sync_names();
            print_matrices(); 
            pw(vectors);
            print_sync_names();
            print_locations(); 
          } 
          if (HALB_OUTPUT) {
              printf(" H -- output \n\n");
              Hprint_var_names(); printf("\n");
              Hprint_matrices();  printf("\n");
              Hprint_vectors();   printf("\n");
              Hpolyhedra_info();   printf("\n");
              printf(" H-output done\n");
            }
          new_automaton();
/*          printf("Parsed %s \n", aut_name);  */
	  ansi_fflush();
	  }
          program
         |
         ;

declarations : variables 
          ;
variables : { var_number = 1; } 
            VARS ':' 
                     {fprintf(stderr,
		     "Old-style var declaration, use 'var x:clock;' etc\n"); 
	             ansi_fflush(stderr);}
	    varlists
          |
            { var_number = 1; } 
            VAR varlists
          ;
varlists : { tmp_var_count = var_count; } vars ':' vartype ';'
           { tmp_var_count = var_count - tmp_var_count;
             add_attribute(yylval.v_type); } varlists
         |
         ;
vartype : INTEGRATOR
        | CLOCK
        | ANALOG
        | PARAMETER
        | DISCRETE
        ;
vars : NAME { var_count++; insert_var((char *)yytext);} vars_next
     |
     ;
vars_next : ',' vars
          |
          ;
sync_vars : NAME { insert_sync((char *)yytext); } sync_vars_next
          |
          ;
sync_vars_next : ',' sync_vars
               |
               ;
synclabels : SYNCLABS ':' { syncs_head = NULL; 
			    /* added HWT */ 
			    sync_number = 1; }
             sync_vars  ';'  
           
/* HWT 2/954 added initialization */
automaton : prolog locations END
          ;
prolog : initialization synclabels
       | synclabels initialization 
initialization : INITIALLY NAME {store_init_loc_name((char *)yytext);
			         rows=0; /* added HTW 5/22 */ 
				 free_vect_list();}
           state_initialization  { get_init_matrix_id(); } 
           ';'
state_initialization: '&'
             linearformula
          |  /* empty */ /* adding, 6/2/95 */
             { rows++; matrix_id++; add_last_vect();
	       add_matrix(); }
          ;
locations :  location locations
          | 
          ;
location : LOC NAME { new_location((char *)yytext); } ':' WHILE
           { rows = 0; /* added HTW 5/22 */ free_vect_list();  }
           linearformula { get_inv_id();  }	
           WAIT '{' {  rows = 0;
	              /* added HWT 2/95 */
	              /* clear_slope_decls(); */
		      /* the original clear slopes */
                      /* clear_slopes();  */ }
           slopvec '}' { 
	                  install_default_slopes();
	                  rows++; matrix_id++; add_last_vect();
			  add_matrix(); 
			  locations->slope_matrix_id = matrix_id;
			 }

/*                          { 
                           get_slopvec();
                           locations->slop_start_vect=vector_id;
                           produce_slope_vectors();
			    /* added HWT 2/95 --
			   /* prev code produced lines instead of rays --
			   /* now fix the slopes into rays --
			   make_slopes_rays(locations->slop_start_vect, 
					    vector_id-1);
			   add_for_undecl_slopes();
                           locations->slop_end_vect=vector_id-1;
                          }
*/
           transitions 
         ;
transitions : { was_asap = 0; new_transition();  } transition
              { locations -> transitions -> was_asap = was_asap; }  transitions
           |
           ;
transition : WHEN  { rows = 0; free_vect_list(); }
             linearformula { get_guard_id();  cur_vec = vector_id; 
			   }
             asssyn
             GOTO NAME { get_destination((char *)yytext); }
             ';'
           ;
asssyn:      non_empty_assignments   { get_assignments(vector_id - cur_vec); }
             synlabel 
           | non_empty_synlabel 
             assignments   { get_assignments(vector_id - cur_vec); }
           | /* neither assignments nor synlabels */
           ;
slopvec    : slop  
           | slop  ',' slopvec
           | 
           ;
slop : slopelinearformula
;
slopelinearterm :  rational     { install_free_term_while(); 
				   if (negated) {
				     sign = -sign;
				     negated = 0;} 
				 }
            | rational NAME { check_slope_name((char *)yytext);
			      install_coef_term_while(slope_name_of((char *)yytext)); 
			      if (negated) {
				sign = -sign;
				negated = 0; } } 

            | NAME          { check_slope_name((char *)yytext);
			      install_coef1_term_while(slope_name_of((char *)yytext));
			      if (negated) {
				sign = -sign;
				negated = 0; } }

            ;
slopelinearexp: slopelinearterm
                | slopelinearterm '+' slopelinearexp
                | slopelinearterm '-' { sign = -sign; negated = 1;} 
                     slopelinearexp
;
slopelinearconstr : slopelinearexp RELOP { 
                               if (yylval.integer == NEQ) { 
				 yyerror("Cannot use != in slope linear expressions");
                               }
                               if (yylval.integer == LT) { 
				 yyerror("Cannot use < in slope linear expressions");
                               }
                               if (yylval.integer == GT) { 
				 yyerror("Cannot use > in slope linear expressions");
                               }
                               install_relop_while(yylval.integer); sign = -1; }
            slopelinearexp
     | slopelinearexp
          IN '[' rational { n1=last_n;d1=last_d; }
        ',' rational { n2=last_n;d2=last_d; }
        ']' 
          { /* need to add two rows, one greater than the first value */
            /* the other less than the second rational */
	    install_slope_bounds(n1,d1,n2,d2);

	  }
/*          { tmp_index = var_index($1.string); 
	    insert_slope(tmp_index, n1,d1,n2,d2 ); 
	      /* added HWT 2/95 --
	      set_slope_declared(tmp_index);} 
*/
          ;

slopelinearformula: { sign = 1; negated = 0; 
		      rows++; new_vect(1 + var_number); }
              slopelinearconstr
               { (void)vector_lcm(); }
             ;

/* ======================================================== */

/* non-empty in sense of a DO is present */
non_empty_assignments: DO '{' {set_assgt_vars_empty();}
                           assignmentlist '}'
            |
             DO '{' '}'
            ;
         

assignments : non_empty_assignments
            |
            ;
assignmentlist : assignment
               | assignment ',' assignmentlist
               ;
 assignment: assgt;


/*
-- originally the following (up till 10/10/95)
assignment : NAME { tmp_index = var_index((char *)yytext); } ASSIGNMENT 
             assignmentspec ;

*/

/* 
  the new jump syntax for transitions, enables arbitrary linear
  expressions among new and old variables, actually supercedes the 
  guard.
  uses assgt, assgt_term, assgt_term1, and assgt_interval
  at first no intervals, just use RELOPs, add intervals later
*/

assgt: { sign = 1; vect_head = NULL; new_vect(1+2*(var_number-1)+1); }
       assgt_exp
       RELOP     { if (yylval.integer == NEQ) { 
	             yyerror("Cannot use != in linear expressions");
                   }
                   install_relop_while(yylval.integer); sign = -1; 
	         }
       assgt_exp { 
	            (void)vector_lcm(); 
		    store_assgt();
		  }
       ;

assgt_term: /* for some reason,  yylval.string doesn't work for  */
            /* "rational" and "rational NAME" */
         rational            { install_free_term_while(); }
       | rational NAME       { install_coef_term_while(yylval.string); }
       | NAME                { install_coef1_term_while(yylval.string); }
       | rational NAME PRIM  { install_coef_p_term(yylval.string); }
       | NAME PRIM           { install_coef1_p_term(yylval.string); }
       ;

assgt_exp: assgt_term
       | assgt_exp '+' assgt_term
       | assgt_exp '-' { sign = -sign; }
         assgt_term    { sign = -sign; }
       ;

/* ================================================== */

/* 7/27. Must change the install to cope with multiple terms w. same var */
/*  or 2 constants */
linearterm :  rational     { install_free_term_while(); }
            | rational NAME { install_coef_term_while((char *)yytext); }
            | NAME          { install_coef1_term_while((char *)yytext); }
            ;
linearexp : linearterm
           | linearexp '+' linearterm
           | linearexp '-' { sign = -sign; } linearterm { sign = -sign; }
           ;
linearconstr : linearexp RELOP {
                               if (yylval.integer == NEQ) { 
				 yyerror("Cannot use != in linear expressions");
                               }
                               install_relop_while(yylval.integer); sign = -1; }
            linearexp
          | P_TRUE   {/* i.e. already set up a vector for this linear */
	              /* expression, now we free it */
		     struct _p_vector * y = vect_head -> next;
                     vect_head -> next = NULL;
                     free_vect_list();
                     vect_head = y;
                     rows--; 
                   }
          | ASAP   { struct _p_vector * y = vect_head -> next;
                     vect_head -> next = NULL;
                     free_vect_list();
                     vect_head = y;
                     rows--;
                     was_asap = 1;
                   }
          | P_FALSE  /* leave it or not ??? */  
                   { /* install a false row,  -1 = 0 */
		       vect_head -> d[0] = EQ;
                       vect_head -> n[0] = EQ;
		       
                       vect_head -> n[var_number] = -1;
	               vect_head -> d[var_number] = 1;

	       
                   }
          ;
linearformula : {
		  sign = 1; rows++; new_vect(1 + var_number); }
                linearconstr 
                { (void)vector_lcm(); }
                linearformularest ;
simple_linearformula: 
                {  /* added HWT 5/22 */ 
                  free_vect_list(); vect_head = NULL; rows = 0;
		  sign = 1; rows++; new_vect(1 + var_number); }
                linearconstr
                { (void)vector_lcm();
		  rows++; matrix_id++; add_last_vect();
		  add_matrix();

		} 
linearformularest : '&' linearformula
                  | { rows++; matrix_id++; add_last_vect();
                      add_matrix();  }
                  ;
integer : INT { $$ = yylval.integer; }
    | '-' INT { $$ = -yylval.integer; }
    ;
rational : integer { last_n = $1; last_d = 1; }
         | integer '/' integer { last_n = $1; last_d = $3; }
         ;
non_empty_synlabel : SYNC NAME { get_synlabel((char *)yytext); }
synlabel: non_empty_synlabel
         |
         ;

/* specification stuff begins here */

specification: old_specification 
          | new_specification
          ;

old_specification : { init_spec_info(); }
            /* if empty spec, assume want reach in POST dirn */
            /* no elimination of locations or variables */
                spec_list
                { old_spec_style = 1;}
              ;

spec_list : single_spec rest_of_spec_list;

rest_of_spec_list: single_spec rest_of_spec_list
            | /* empty */
            ;

single_spec: dirn_spec
          |  eliminate_locns_spec
          | eliminate_vars_spec
          | final_spec
          ;
dirn_spec : DIRECTION ':' P_PRE { assign_pre_dirn(); }
         |
            DIRECTION ':' P_POST {assign_post_dirn(); }
         ;

eliminate_locns_spec : 
            ELIM_ALL_LOCNS ':' P_TRUE { assign_elim_locns(1); }
         |
            ELIM_ALL_LOCNS ':' P_FALSE { assign_elim_locns(0); }
         | ELIM_LOCNS ':' {init_elim_locns();}
           NAME {add_elim_proc((char *)yytext);} rest_elim_proc_list
           ';'
         ;
rest_elim_proc_list: 
          ',' NAME {add_elim_proc((char *)yytext);} rest_elim_proc_list;
         | /* empty */
         ;

eliminate_vars_spec :
            /* ELIM_ALL_VARS really means all non-parameter vars */
           ELIM_ALL_VARS ':' P_TRUE { assign_elim_vars(1); }
         |
            ELIM_ALL_VARS ':' P_FALSE { assign_elim_vars(0); }
         | ELIM_VARS ':' {init_elim_vars();}
           NAME {add_elim_var((char *)yytext);} rest_elim_var_list
           ';'
         ;
rest_elim_var_list: 
          ',' NAME {add_elim_var((char *)yytext);} rest_elim_var_list;
         | /* empty */
         ;
         
final_spec: FINAL ':' {init_pred_spec(); }
         final_spec_list
            {assign_final_spec(); }
          ;
final_spec_list:  {init_convex_predicate();} 
                 convex_spec 
                 {assign_convex_spec();}
                 final_spec_rest
                
                 ;
final_spec_rest: 
		  '|' final_spec_list
         | /* empty */
         ;
convex_spec:  /* a non-empty conjunct */
         single_convex_spec convex_spec_rest
  ;

convex_spec_rest:
         '&' convex_spec  
         | /* empty */
         ; 
single_convex_spec: /* either a state pred part or a variable part */
         '(' single_convex_spec ')'
         |
         simple_linearformula 
           /* {fprintf(stderr,"Got formula\n"); ansi_fflush(stderr);} */
         |
          /* beware this also allows ASAP? */
          /* it also adds the matrix to the global list */
         state_pred 
          /* {fprintf(stderr,"Got state pred\n");ansi_fflush(stderr);} */
         ;
state_pred: /* can give equality only */
       LOC
       '[' 
       NAME  {state_aut_name = str_preserve((char *)yytext);}
       ']'  
       RELOP /* really want eq here, but it's lexically analyzed as a RELOP */
       { if (yylval.integer != EQ) { 
	   yyerror("Syntax error in location expression");
         }
       }
       NAME 
       { insert_state_predicate(state_aut_name,(char *)yytext);}
         ;

/* =================== new specification stuff ============== */

new_specification:
       { /* init_region_declarations(); */ }
       region_declarations
       {
	 /* fprintf(stderr,"done with region declns\n");ansi_fflush(stderr);*/
       init_statements(); }
       statements
       {global_stmt = stmt_stack[stmt_stack_nr];
/*
	fprintf(stderr,"We got stmt_depth = %d\n", stmt_stack_nr);
	print_statements(global_stmt);
	fprintf(stderr,"done with statements\n"); 
	ansi_fflush(stderr);
*/
        old_spec_style = 0; }
       ;

region_declarations:
       VAR {init_region_name_table();}
	 region_name_list ':' REGION ';'
	   |
    ;  /* empty --- no declarations */

region_name_list:
        region_name_decl 
        rest_of_region_name_list
        ;
    
region_name_decl:
       NAME {store_region_name((char *)yytext);
		/* fprintf(stderr,"declared with region: %s\n", (char *)yytext); */
	   }
       |
       REGNAME {fprintf(stderr,
			 "Warning: redeclaration of region %s ignored\n",
			(char *)yytext);
		 ansi_fflush(stderr);
	       }
        ;
rest_of_region_name_list:
        ',' region_name_list
        | /* empty */
        ;

/* after statements parsed, head of list is in stmt_stack[stmt_stack_nr] */
/* MUST call init_statements(); before starting to parse statements */
statements: statement_list  {store_statements();}
            ;

statement_list: /* allow empty statements? why though? Not now. */
              statement ';'  
                /* {print_statement(stmt_stack[stmt_stack_nr]);} */
              restofstatements 
                            ;

restofstatements: statement_list
    | /* empty */
    ;

/* after parsing a statement, it's is stmt_stack[stmt_stack_nr] */
statement:
    FREE REGNAME {init_statement(FREE_STMT);
		  store_name_in_stmt((char *)yytext);
		  /* fprintf(stderr,"freeing %s\n",(char *)yytext);
		     ansi_fflush(stderr);
		  */
		}
 |
    PRINT OMIT  {init_statement(PRINT_SIMPLIFIED_STMT);
		 init_elim_loc_info();}
            elim_location_info { store_elim_loc_info(); }
	    LOCATIONS
            region_expression  
                    {
		    /*
		    fprintf(stderr,"printing %s\n",(char *)yytext);
		    ansi_fflush(stderr);}
                    */
		    store_expr_in_stmt();
		  }
    |
    PRINT           {init_statement(PRINT_STMT); }
            region_expression  
                    {
		    /*
		       fprintf(stderr,"printing %s\n",(char *)yytext);
		       ansi_fflush(stderr);}
                    */
		    store_expr_in_stmt();
		  }
   |
    PRINTSIZE     {init_statement(PRINTSIZE_STMT); }
            REGNAME  
                    {
		    /*
		       fprintf(stderr,"printing %s\n",(char *)yytext);
		       ansi_fflush(stderr);}
                    */
		    store_name_in_stmt((char *)yytext);
		  }
   |
    PRINTSTRING STRING  {init_statement(PRINT_STR_STMT);
		    store_outputstr_in_stmt(yylval.string);
		    /* fprintf(stderr,"printing string %s\n",yylval.string);
		       ansi_fflush(stderr);
                    */
                    }
    |
    PRINT TRACE TO {init_statement(TRACE_STMT); }
    region_expression {store_expr_in_stmt();}
    USING
    REGNAME {store_name_in_stmt((char *)yytext);}
    |
    REGNAME {init_statement(ASSIGN_STMT);
	     store_name_in_stmt((char *)yytext);
	     /*
		fprintf(stderr,"assigning %s\n",(char *)yytext);
		ansi_fflush(stderr);`
	     */
	   }     
    ASSIGNMENT
    region_expression {store_expr_in_stmt();}
    |
    IF Bool_expression { init_statement(IF_THEN_STMT); 
			 store_Boolexpr_in_stmt();
		         init_statements(); }
    THEN statements    { store_stmt_in_stmt(IF_THEN_STMT); }
    elsepart
    | 
    WHILE Bool_expression DO { init_statement(WHILE_STMT);
			       store_Boolexpr_in_stmt();
			       init_statements(); }
     statements              { store_stmt_in_stmt(WHILE_STMT); }
    ENDWHILE
    ;

elsepart:
    ENDIF
    |
    ELSE { init_statements(); }
    statements     {  store_stmt_in_stmt(IF_THEN_ELSE_STMT); }
    ENDIF
    ;

/* after parsing a region expression, the expr is in expr_stack[expr_depth] */
region_expression: 
/*     {init_expression(OR);}  */
    region_expression 
    '|'    /* P_OR */ { /* fprintf(stderr,"OR\n"); */ } 
    conj_region_expression {store_or_expression();} 
   |
   conj_region_expression
   ;

conj_region_expression:
	conj_region_expression  '&'  /* P_AND    */
       {/* fprintf(stderr,"AND\n"); */ }
	basic_region_expression  
	 {store_and_expression();}
    |
    basic_region_expression
    ;

basic_region_expression:
  
   P_WEAKDIFF 
   '('
   region_expression ',' region_expression 
   ')' {store_weakdiff_expression();}
   |
   P_DIFF 
   '('
   region_expression ',' region_expression 
   ')' {store_diff_expression();}
   |

   REGNAME 
   {init_expression(REGION_NAME);
    store_name_expression((char *)yytext);
    /* fprintf(stderr,"expr: %s\n",yytext); */
   }
    |
    '~' {init_expression(COMPLEMENT);}
    basic_region_expression
    {store_first_expression();}
    |
    '(' { /* fprintf(stderr,"( ");ansi_fflush(stderr); */ }
       region_expression ')'  
      {/* fprintf(stderr,")\n ");ansi_fflush(stderr); */ }
    |
	P_PRE '('  {init_expression(PRE_EXPR); 
		    /* fprintf(stderr,"PRE ( "); */
		   }
       region_expression ')' 
      {store_first_expression(); 
       /* fprintf(stderr,")\n "); */
      }
    |
	HULL '('  {init_expression(HULL_EXPR);
		   /* fprintf(stderr,"HULL ( "); */
		  }
       region_expression ')'  {store_first_expression();
			       /* fprintf(stderr,")\n "); */
			      }
    |
    P_POST '('  {init_expression(POST_EXPR); 
		 /*fprintf(stderr,"POST ( "); */
	        }
       region_expression ')'  {store_first_expression();
			       /* fprintf(stderr,")\n "); */
			      }
    |
    HIDE {init_expression(HIDE_EXPR); init_elim_var_info(); }
    elim_variable_info { store_elim_var_info(); }
    IN region_expression {store_first_expression();}
    ENDHIDE
    |
    REG_OMIT  {init_expression(OMIT_EXPR);
		 init_elim_loc_info();}
            elim_location_info { store_elim_loc_info(); }
	    LOCATIONS
            region_expression  
                    {
		    store_first_expression();
		  }
    |    REACH FORWARD FROM {init_expression(REACHFORWARD); 
			/* 
			   fprintf(stderr,"reach forward from ( ");
			   ansi_fflush(stderr);
			*/
		        }
       region_expression ENDREACH {store_first_expression();
				   /*
				   fprintf(stderr,")\n ");
				   ansi_fflush(stderr);
				   */
				   }
    |
    REACH BACKWARD FROM {init_expression(REACHBACKWARD);
			 /* fprintf(stderr,"reach forward from ( ");*/
		        }
       region_expression {store_first_expression();
			  /* fprintf(stderr,")\n "); */
			}
      ENDREACH 
     |
	ITERATE REGNAME {init_expression(RECURSION);
			 /*
			    fprintf(stderr,"depth rec expr = %d\n ",expr_depth);
			    ansi_fflush(stderr);
			 */
			 store_name_in_expression((char *)yytext);} 
        FROM 
         region_expression {store_first_expression();} 
	USING '{' {init_statements();}
        statements '}' {
	                /* fprintf(stderr,"Hi\n");
			   fprintf(stderr,"depth rec expr = %d\n ",expr_depth);
			   ansi_fflush(stderr);
			*/
			store_stmts_in_expr();} 
   | 
    /* just a conjunct */
    region_predicate_short {init_expression(PREDICATE); 
		      store_convex_pred_expression();}
   ;

 region_predicate_short:
    { init_spec_info(); init_pred_spec(); init_convex_predicate();}
/* was convex spec,. but this creates parser problems with x<=0 & regionname */
    single_convex_exp 
    {assign_convex_spec();}
         ;

single_convex_exp: /* either a state pred part or a variable part */
         simple_linearformula 
                /* {fprintf(stderr,"Got formula\n"); ansi_fflush(stderr);} */
         |
          /* beware this also allows ASAP? */
          /* it also adds the matrix to the global list */
         state_pred 
                /* {fprintf(stderr,"Got state pred\n");ansi_fflush(stderr);} */
         ;


 elim_variable_info:
        ALL {assign_elim_vars(2); }
        |
        NON_PARAMETERS {assign_elim_vars(1); }
        |
        {init_elim_vars();}
        NAME {add_elim_var((char *)yytext);} 
        rest_elim_var_list
        ;
 elim_location_info:
        ALL { assign_elim_locns(1); }
        |
         {init_elim_locns();}
        NAME {add_elim_proc((char *)yytext);} 
        rest_elim_proc_list
        ;

 Bool_expression:
   Bool_expression
   P_OR
   conj_Bool_expression {store_or_Boolexpression(); }
   |
   conj_Bool_expression
   ;

 conj_Bool_expression:
   conj_Bool_expression
   P_AND
   basic_Bool_expression  {store_and_Boolexpression(); }
   |
   basic_Bool_expression
   ;

 basic_Bool_expression:
   region_expression 
   RELOP             {$2 = yylval.integer;}
   region_expression {init_Boolexpression(BE_INEQ);
		      store_ineq_type_in_Boolexpression($2);
		      store_ineq_Boolexpression();}
   |
/*
   region_expression 
   P_LT              
   region_expression {init_Boolexpression(BE_INEQ);
		      store_ineq_type_in_Boolexpression(LT);
		      store_ineq_Boolexpression();}
   |
   region_expression 
   P_GT              
   region_expression {init_Boolexpression(BE_INEQ);
		      store_ineq_type_in_Boolexpression(GT);
		      store_ineq_Boolexpression();}
*/
   |
   region_expression 
   WEAKLE             {$2 = yylval.integer;}
   region_expression {init_Boolexpression(BE_WEAKLE);
		      store_ineq_Boolexpression();}
   |
   region_expression 
   WEAKEQ            
   region_expression {init_Boolexpression(BE_WEAKEQ);
		      store_ineq_Boolexpression();}
   |
   region_expression 
   WEAKGE             
   region_expression {init_Boolexpression(BE_WEAKGE);
		      store_ineq_Boolexpression();}
   |
   P_NOT
   basic_Bool_expression {store_not_Boolexpression();}
   |
   EMPTY
   '('
   region_expression {store_expr_in_empty_Boolexpression();}
   ')'
 
/*
   |
   P_TRUE {init_Boolexpression(BE_TRUE);
	 }
|
   P_FALSE {init_Boolexpression(BE_FALSE);
	 }
*/
  ;

%%


int
yyerror(const char *yerr_str) {
  fprintf(stderr,"Error message: %s\n", yerr_str);
  fprintf(stderr,"Error in input on line %d while scanning input: %s\n",
	  yylineno, (char *)yytext);
  exit(1);
}


     
      
	       
