#ifndef __INCLUDE_H
#define __INCLUDE_H

#include <stdio.h>
#include <stdlib.h>   /* moved here 4/26/96 */
#include <malloc.h>
#include <string.h>


#ifdef __cplusplus
extern "C" {
#endif

/* added HWT 12/95 for option in using library routines */
#define FAST 0
#define SLOW 1
extern int overflow_avoid_mode;

extern int poly_routine_id;

#include "fflush.H"

/* HWT: beware, the exact values chosen are essential, since they're used
  in the first entry of matrices, i.e. EQ=0, GE=1, GT=2 */
enum { EQ = 0, LE = 5, GE = 1, NEQ = 3, LT = 4, GT = 2 };       /* types of operators */
enum variable_type { V_INTEGRATOR=0, V_CLOCK=1, V_PARAMETER=2, V_ANALOG=3, V_DISCRETE=4}; /* types of variables */

/* some global variables */
extern int rows, columns, matrix_id, tmp_index, cur_vec, sl1, 
    tmp_var_num, var_number, sync_number, loc_number, some_slope,
    tmp_var_count, trans_num, was_asap;

/* values of the follwoing are instantiated in include.c */
extern int var_count,  IN_SLOPE;

extern int SIMPLE_OUTPUT, HALB_OUTPUT; /* output fromat */                   

/* variable names data structures + stuff for names of synclabels/locations */

struct var_name {
  char * name;
  int index, attribute;
  struct var_name * next;
} ;

extern struct var_name *var_head, *syncs_head, * tmp_var_head;

int var_index(char * name), var_type(int index), sync_index(char * name);
void insert_var(char * name), insert_sync(char * name), print_var_names(void);
void add_attribute(int attr); 
void free_var_list(void);
void free_var_list(void);
void add_matrix(void);
void free_vect_list(void);
void free_vectors(void);
void free_matrix_list(void);
void new_vect(int len);
void install_free_term_while(void);
void install_coef_term_while(char * name);
void install_coef1_term_while(char * name);
/* next two routines for assignment terms */
void install_coef_p_term(char *name);
void install_coef1_p_term(char *name);

void install_relop_while(int optype);
void install_slope_bounds(int lown,int lowd,int upn, int upd);
int gcd(int x, int y), lcm(int x, int y);
int vector_lcm(void);
void vect_clear(void);
void add_last_vect(void);
void transform_assignment(int var_id);
/* added HWT 3/95. for assignment to single term */
void transform_single_assignment(int var_id);

/* added HWT 10/95 */
/* stores a single assignment constraint */
void store_assgt();

void free_slope_info(void);
void clear_slopes(void);
void slope_init(void);
void insert_slope(int index, int n1, int d1, int n2, int d2);

/* added HWT 3/95, for unbounded assignments */
void add_minus_inf_ass();
void add_pos_inf_ass();

void clear_slope_decls();
void add_for_undecl_slopes();
void set_slope_declared(int indexnr);

/* for more numerical things ... */

extern int last_n, last_d; /* nominator & denominator of last seen rational */
extern int sign;           /* used for multiplying when term found to the
                       right of RELOP */
extern int vect_len;       /* length of currently prepared vector */

struct _p_vector {
  int len;
  int vect_id;
  int variable; /* used when assignment to a variable */
  int * d, * n; /* nominators & denominators */
  int vlcm;     /* vector's lcm used to scale to integers, for slopes only */
  struct _p_vector * next;
};

extern struct  _p_vector * vect_head, * vectors;
/* vect_head is used by routines that manipulate vectors;
   vectors is a list of vectors used in assignments */

struct _p_matrix {
  int matrix_number;
  int columns, rows;
  struct _p_vector * vectors;
  struct _p_matrix * next;
};

extern struct _p_matrix * matrix_head;

extern int vector_id;
extern int DEFAULTS[5];
  
struct slope_entry {
  int * min_d, * min_n, * max_d, * max_n;
  int len;
} ;

extern struct slope_entry * slope_info;

extern int n1,d1,n2,d2; /* for YACC internal use */


/* HWT for initialization */
/* added global declarations, 2/95 */
extern int init_matrix_id;  
extern char *init_location_name;
extern int * slope_delared;   /* array of booleans for whether slope has */
                              /* been declared at this location */
extern int reach_dirn;        /* dirn in which to do reach analysis */
/* all these actually want booleans */
extern int elim_all_locns;
extern int elim_all_vars;
extern int had_dirn_spec;
extern int had_elim_l_spec;
extern int had_elim_v_spec;
extern int had_final_spec;
extern int elim_locs_by_list;
extern int elim_vars_by_list;
extern struct var_name *elim_proc_list;
extern struct var_name * elim_var_list;

extern int negated; /* for use in slopelinearformulas */

extern char *aut_name;  /* name of the automaton (component) */

extern int *assgt_vars;


void store_automaton_name(char *autname);

/* for error mesage, then exiting */
void
yyerrormessage(char *errorstr);

/* just a warning */
void
yywarningmessage(char *warningstr);

/* returns 0 if not equal, 1 if equal */
/* equality means exact same coefficients (n and d) */
int p_vector_eq(struct _p_vector *vec1, struct _p_vector *vec2);

/* check that no fixed slope var names are used in slope constraints */
/* check the name is a declared var, with dvarname having 'd' in front */
void check_slope_name(char *dvarname);

char *
slope_name_of(char *dvarname);



#ifdef __cplusplus
}
#endif
#endif


