/*
  Changes from version 1.02b

  Get rid of all the closure difference stuff.
     Do exact difference.  Add declarations for opoly::complement
     in opoly.h 
   Change input syntax to allow operator diff(region-var,region-var) instead 
     of cldiff
   added field autp to classes region_array and location.  It points to basis 
      automaton.  e.g. for complementing
      empty lists, need to know the locations for the full regions.
      
  add NOT operator on regions to input syntax as the tilde (~) symbol.
  remove print statements in REACHBACKWARD (in exec_spec.C)
  added error checking on array bounds on expression and stmt stacks in spec.c

--

patches to v1.04

12/5/96

* BUG FIX: added constructor region_array(int sz, automaton* automaton_p)
fixed constructor calls in ra_complement, and in ra_reach for objects
temp_ra and succ_ra

* fixed bug in region_array::is_empty, used to consider first element of
  each list only, instead of entire list.

* freed arg_compl in r_a::ra_minus(), comp_of_single in region::complement_list()

* fixed bug. in ra_minus, check for empty list first

* fixed bug in Post/ Pre commands. added restriction to invariant in exec_spec.C.  also for efficiency, added simplify() step after finding Pre()

* fixed bug in region_array::restrict_to_invariant(), checks for
non-empty region lists first.

------

fixes for port to Linux

6/2/00.

main compilation of hytech.exe uses lex compatibility mode of flex in
Makefile-linux.

in include.h, declared type of var_count and IN_SCOPE to be int, i.e.,
"extern int var_count, IN_SCOPE"

Plugged a memory leak: in extrapoly.C, in ok_to_generate(), line 863,
changed type (and name) of pt_poly and gen_poly to pointers
pt_poly_ptr and gen_poly_ptr, etc.

fixed "typedef char bool" to "typedef char Bool", i.e., declaring a
new type "Bool", to avoid conflict with predefined "bool" class in C++

8/14/00. Bug fix: in spec.c, in store_expr_in_empty_Boolexpression() 
      correctly incremented ptr into Bool expression stack before 
      assignment of expression.
      i.e., added the "Boolexpr_depth++" before 
      "Boolexpr_stack[Boolexpr_depth] = Boolexpr;"

9/13/00. Added type definition for COEFF, the type for the
      coefficients in the matrices representing polyhedra.
      set via the user defining either COEFF_IS_INT or 
         COEFF_IS_LONG_LONG_INT in the Makefile.

       Added files basic_cpp.[CH] (for clear_coeff_array() )

9/13/00.  incorporating a fix from 1/15/97 from another 
            code stream:
	    in compose, locations' autp field was not 
	    previously set,  corrected in 
	    automaton::set_up_reach_locns();  however, the 
	    copying in automaton_info's reach_compose and 
	    compose stuff up the autp ptr values, so these 
	    are reset in these routines.

9/14/00.    added opoly::copy_for_slopes(),  a plain copy that 
              doesn't check emptiness, etc, of slopes.

	      v1.04f
11/17/00. fixed foutput() in region.C.

6/8/01. added lib_polyhedron::contains_origin()
        changed argument of ok_to_generate() from frame to lib_polyhedron
        checks for origin in slope poly (previously only checked for
          the origin as a point in the frame representation)

8/10/01 - 9/12/01.
. fixed time_evolving to allow for a frame matrix containing just the
 origin, for which fixed range over potential rows to include the final row

9/12/01.
* added line numbers for error output for undeclared variables
* added code for reg_omit

9/20/01. changed code to abort on duplicated variable declaration, and
     duplicated listing of a sync label.

11/30/01. increased MAX_VARS to 150.

12/19/01. in cherni.C, added use of myadd in place of old myinc
added mymult usage to slope_poly()

12/24/01. fixed type of b0, b in ori_translate, and prod in ori_restore
     from 'int' to 'coeff'. This had lead to non-detected errors in integer
     computations, caused by truncating a "long int" to an "int" when
     creating constraint matrices from frame matrices.

1/24/02.    changed region::and to region::region_and 
     alterations required in region.H, and diff.C only.

*/



#include <stdio.h>

#include <string.h>
#include <stdlib.h>
// #include <stream.h>

  // i'm going to declare something.
  //  test to see whether the C data structures can be accessed
  //  attempt to DECLARE C linkage


#include "include.h"
#include "location.h"
#include "output.h"
#include "parser.h"
#include "spec.h"
#include "exec_spec.H"
#include "getresources.h"

extern "C" {
extern FILE *yyin;
extern void lexinit();
extern void yyparse();

} // end of C linkage header files


#include "main.H"

#define COMPOSE_NO_REACH 0
#define COMPOSE_REACH 1

int compose_mode = COMPOSE_REACH;


void
outputline() {
  cout << 
     "=================================================================\n" ;
} 


void
out_of_memory() {
  cout << "ABORTING\n";
  print_max_mem();
  cout << "Program aborted. Out of memory\n";
  exit(1);
}

// currently allow only -p0, and -p1 for printing not much, and lots
//   -p2 prints regions and iteration "."'s.
void
process_flag(char *flagstr) {
  switch (flagstr[0]) {
  case 'i':
    // output more info, then quit

     cout << "Version 1.04f (last modified 9/20/01) from v1.04a of 12/6/96\n";
     cout << "For more info: \n";
     cout << "    email: hytech@eecs.berkeley.edu\n";
     cout << "    http://www.eecs.berkeley.edu/~tah/HyTech\n";
     outputline();
     cout << "HyTech can be compiled with coefficients in polyhedra \n";
     cout << "      represented by either type 'int' or 'long long int.'\n";
     cout << "      This version compiled using type ";
#ifdef __COEFF_IS_LONG_LONG_INT
       cout << "'long long int.'";
#else
     cout << "'int.'";
#endif
     cout  << "\n";
     outputline();
    cout << "Input format has changed since 10/31/95\n";
    cout << "   Rate declarations: use dx instead of x' for dx/dt, dy instead of y', etc\n";
    cout << "   Assignments: arbitrary linear expression between old and new variable\n";
    cout << "      values are allowed. Use x to refer to value of variable x before\n";
    cout << "      the transition takes place, and x' to refer to its new value\n";
    cout << "      If the new value is not specified, it is assumed to be the same\n";
    cout << "      as the old value\n";
    cout << "      Also, -inf, and inf are no longer permitted\n";
    cout << "      Hence, replace x := [3,6] with 3<=x', x'<=6\n";
    cout << "             replace x := (3,inf) with x'>3\n";
    cout << "             replace x := (-inf,inf) with x'=x'\n";
    cout << "\n";
    cout << "Versions 1.02(a,b), 1.04 also allow strict inequalities in invariants, guards,\n";
    cout << "      and assignments, but NOT in rate conditions\n";
    cout << "For latest options, type 'hytech'\n";
    cout << flush;
    exit(1);
    break;
  case 'p' : 
    // degree of Printed output, higher is more
    if ((strlen(flagstr) < 2) || (flagstr[1] != '0') && (flagstr[1] != '1')
                                && (flagstr[1] != '2')) {
      cerr << "Warning: illegal flag: -p0, -p1, or -p2 required\n" << flush;
    } 
    else {
      print_mode = atoi(flagstr+1);
      cout << "Setting output level to " << print_mode << "\n";
    }
    break;
   case 'f' : 
    // formatting mode, 0 = horizontal, 1 = vertical
     if ((strlen(flagstr) < 2) || (flagstr[1] != '0') && (flagstr[1] != '1')) {
       cerr << "Warning: illegal flag: -f0, -f1 required\n" << flush;
     } 
     else {
       format_mode = atoi(flagstr+1);
       cout << "Setting format mode to " << format_mode << " (" <<
	 ((format_mode ==1) ? "vertical" : "horizontal") << ")\n";
     }
    break;
  case 'c' :
    // whether to check for consistency, 0 = no, 1 = yes
     if ((strlen(flagstr) < 2) || (flagstr[1] != '0') && (flagstr[1] != '1')) {
       cerr << "Warning: illegal flag: -c0, -c1 required\n" << flush;
     } 
     else {
       check_automaton_flag = (flagstr[1] == '1');
       cout << "Will " << ((check_automaton_flag) ? "" : "not ") << 
	 "check automata for consistency\n";
     }
    break;
   case 'o' :
    // whether to divide by GCD to avoid overflow errors in 'cherni.C::combine'
     if ((strlen(flagstr) < 2) || (flagstr[1] != '0') && (flagstr[1] != '1') 
	 && (flagstr[1] != '2')) {
       cerr << "Warning: illegal flag: -o0, -o1 required\n" << flush;
     } 
     else {
       switch (flagstr[1]) {
       case '0': 
	 overflow_avoid_mode = 0;
	 cout << "Will not try to avoid library arithmetic overflow errors\n";
	 break;
       case '1': 
	 overflow_avoid_mode = 1;
	 cout << "Will try to avoid library arithmetic overflow errors\n";
	 break;
       case '2': 
	 overflow_avoid_mode = 2;
	 cout << "Will try hard to avoid library arithmetic overflow errors\n";
	 break;
       default: overflow_avoid_mode = 0;
       }
      }
    break;
 case 'r' :
    // whether to use control reachability in composition, 0 = no, 1 = yes
     if ((strlen(flagstr) < 2) || (flagstr[1] != '0') && (flagstr[1] != '1')) {
       cerr << "Warning: illegal flag: -r0, -r1 required\n" << flush;
     } 
     else {
       compose_mode = (flagstr[1] == '1'); // COMPOSE_REACH == 1
       cout << "Will " << ((compose_mode) ? "" : "not ") << 
	 "generate only control-reachable locations in composition\n";
     }
    break;


  default: 
     cerr << "Warning: illegal flag (" << flagstr << ") ignored\n";
     break;
  }
}


void
process_flags(int agc, char *agv[]) {
// set control variables for verification.
// assume agv[0] is name of program.
// agv[agc] is input file name if it isn't a flag.
// 
// e.g. print_mode = 0 for not much, = 1 for give me heaps
  // don't do anything for now.
  int i;

  // set defaults
  print_mode = 0;                 // quiet
  format_mode = 0;                // horizontal
  check_automaton_flag = true;

  if (agc==1) {
    // just had 'hytech'  only on command line, no filename
    // display options instad of analyzing hybrid system
    cerr << "HyTech: version 1.04f\n";
    cerr << "Tool for symbolic analysis of hybrid systems\n";
    cerr << "Usage: hytech [-<flag><n>]* <filename> \n";

    cerr << "  Options: \n";
    cerr << "                [def] indicates default\n";
    cerr << "           -c   consistency checks of automata\n";
    cerr << "                0:no, 1:yes [def]\n";
    cerr << "           -f   formatting of output\n";
    cerr << "                0:horizontal [def], 1:vertical\n";
    cerr << "           -i   more info on this version of HyTech\n";
    cerr << "           -o   overflow avoidance\n";
    cerr << "                0-2: 0:no avoidance [def], 1:some 2:maximal\n";
    cerr << "           -p   level of output\n";
    cerr << "                0: minimal [def], 1:normal, 2:verbose\n";
    cerr << "           -r   perform (untimed) reachability during composition\n";
    cerr << "                0:no, 1:yes [def]\n";

  }
  else {
    for(i=1;i<agc;i++) {
      if (agv[i][0] == '-') {
	// then process a flag
	process_flag(agv[i]+1);
      }
      else {
	if (i != agc-1) {
	  cerr << "Warning: illegal flag: " << agv[i] << "\n" << flush;
	}
	// else last argument is the filename
      }
    }
  }

}

void
intro_output() {
  outputline();
  cout << "HyTech: symbolic model checker for embedded systems\n";
  cout << "Version 1.04f (last modified 1/24/02) from v1.04a of 12/6/96\n";
  cout << "For more info: \n";
  cout << "    email: hytech@eecs.berkeley.edu\n";
  cout << "    http://www.eecs.berkeley.edu/~tah/HyTech\n";
  cout << "Warning: Input has changed from version 1.00(a). Use -i for more info\n";
  outputline();
  cout << "\n";
}

//  increment the dimension of the global state space by adding a column
// for the eps_slack variable.
void
increment_dimension(int *global_varcount ) {
  *global_varcount = var_count+1;
  // insert_var updates the list of var names in var_head, and also var_number
  //   which tracks IDs for variables
  insert_var("eps_slack");
  // (treat eps as a parameter. this way it'll do the right thing when
  //    existentially quantifying all non-parameters, too).
  var_head->attribute = V_PARAMETER;
}


// ==================================================



void
old_spec_verify(automaton *aut) {

  cout << "========================================\n";
  cout << "Attempting verification\n";
  cout << "========================================\n";

  cout << "In direction: " << ((reach_dirn ==PRE) ? "PRE" : "POST") ;
  cout << "\n" << flush;

   /* reach_dirn is parsed in , during the spec section of the input file */
  region_array *reachedarr;
  region_array *final_arr;
  if (reach_dirn == 0) {
    reachedarr = aut->new_reach(PRE);
  }
  else {
    reachedarr = aut->new_reach(POST);
  }
  if (reach_dirn == 0) {
    // for PRE
    final_arr = final_region_array(aut,PRE);
  }
  else {
    // for POST, we want the final region, it's same as initial in PRE dirn
    final_arr = final_region_array(aut,POST);
  }

  region_array *randfinal = reachedarr->ra_and(final_arr);
  outputline();
  cout << "Intersection of reachable with final regions\n";


  if (randfinal != NULL) {
    randfinal->print();
  }
  outputline();
  cout << "Simplified output\n" << flush;
  region * resultreg = randfinal->output_elim();

  // test of trace generation
  reachedarr->generate_trace_to(final_arr);


}


// execute the specs for the given automaton.  
void
execute_specifications(automaton *aut, struct statement *stmt, 
		       struct region_name_table *reg_nam_table) {
   struct region_table * reg_table = 
     (struct region_table *) malloc (sizeof(region_table));
   set_up_reg_table(reg_table,reg_nam_table);
   execute_statements(stmt,reg_table,aut);
   cout << "\n\n";
}

int
main( int argc, char *argv[] ){
  automaton a;  // a C++ automaton
  automaton prod_a; // product automaton
  automaton_info ainfo; // a list of C++ automata
  region_array reached_set;
  // the C automata are parsed into the C structure called 'automata'
  ios::sync_with_stdio();
  set_new_handler(out_of_memory);

  intro_output();
  process_flags(argc,argv);
  // terminate if only display options, no arguments, i.e. no filename
  if (argc==1) {
    exit(1);
  }

  infilename = argv[argc-1];  
  if( (argc >= 2) && (argv[argc-1][0] != '-') ) {
    yyin = fopen( argv[argc-1], "r" );	/* doesn't deal with recursive loads */
    if (yyin==NULL) { 
      fprintf( stderr, "*** Error opening `%s' ***\n", argv[argc-1] );
      fflush(stderr);
      yyin=stdin;
    }/*  else printf( "--Taking input from %s...\n", argv[argc-1] ); */
  }
  else {
    fprintf( stderr, "*** Error: no input file given **\n");
    printf("*** Error: no input file given **\n");
    exit(1);
  }


  // C++ lib seems to want these functions to be defined with __Fv() )
  lexinit();
  yyparse();


  // this constructor parses a C list into a C++ list
  ainfo = automaton_info(*automata);

  // from this point on, we use var_count to refer to n+1, i.e. include eps
  // as a variable
  increment_dimension(&var_count);
  make_var_names();


//  automaton_info *prodp = ainfo.compose();

  automaton_info *prodp;
  if (compose_mode == COMPOSE_REACH) {
    prodp = ainfo.reach_compose();

  }
  else {
    // do the full composition
    prodp = ainfo.compose();
  }

// prodp->a->print();

  // dispose of this routine altogether later
  if (!test_ok_to_generate(prodp->a)) {
    cerr << "Illegal rate conditions: program aborting\n" << flush;
    exit(1);
  }


  make_exist_vectors(var_count,var_count);
    
  prodp->a->set_elim_info();  

  if (old_spec_style) {
    cerr << "Using old specification style\n" << flush;
    if (reach_dirn ==0) {
      prodp->a->set_pre();
    }
    old_spec_verify(prodp->a);
  }
  else {
    execute_specifications(prodp->a,global_stmt, &reg_name_table);
  }
  print_max_mem();
  cout << flush;
  cerr << flush;

  return (0);
}


