#include "output.h"

/* must put global vars here, not in header where they are declared extern */
int matrix_print;

void print_vars(x)
struct var_name * x;
{
  if (!x) return;
  print_vars(x -> next);
  printf(" %s/%d", x -> name, x-> index);
}

void print_var_names() 
{ printf("names: ["); print_vars(var_head); printf(" ]\n"); }

void print_sync_names()
{ printf("syncs: ["); print_vars(syncs_head); printf(" ]\n"); }


/* 'pretty_print' -- only for diagnostic purposes */
void print_var_names1()
{
  struct var_name * x = var_head;

  printf("Number of variables : %d\n", var_count);
  for (; x; x = x -> next) {
    printf("index:%d VAR_NAME::%s/", x->index,x->name);
    switch (x -> attribute) {
      case V_PARAMETER: printf("parameter\n"); break;
      case V_DISCRETE: printf("discrete\n"); break;
      case V_CLOCK: printf("clock\n"); break;
      case V_ANALOG: printf("analog\n"); break;
      case V_INTEGRATOR: printf("integrator\n"); break;
      default: break; 
    }
  }
}

/* printing vectors in reverse order */
void pw(x)
struct _p_vector * x;
{
  int i;
  if (x == NULL) return;
  pw(x -> next);
  if (x -> variable != -1)
    printf("VectID:%d VarID: %d ", x->vect_id,x->variable);
  else if (!matrix_print) printf("VectID:%d  ", x->vect_id);
  printf("[");
  for (i = 0; i < x -> len; i++)
      printf("%3d  ", x -> n[i]);
    printf("]\n");
}

void print_vect()
{
  printf("%d (%d,%d)\n", matrix_id, rows, columns);
  pw(vect_head);

}

/* print matrices in reverse order */
void pm(x)
struct _p_matrix * x;
{
  if (!x) return;
  pm(x -> next);
  printf("%d: (%d %d) [\n", x->matrix_number, x->rows,x->columns);
  matrix_print = 1;
  pw(x -> vectors);
  matrix_print = 0;
  printf("]\n");
}

void print_matrices()
{
  printf("matrices: %d\n", matrix_id);
  pm(matrix_head);
}

void print_loc_slopes(loc)
struct info_location * loc;
{ 
  printf("SLOP-VECTORS:<%d,%d>\n", loc->slop_start_vect,
                                   loc->slop_end_vect); 
  printf("Matrix ID for slopes %d\n",loc->slope_matrix_id);
}

void print_assignments(x)
struct info_assign * x;
{
  if (x == NULL) return;
  print_assignments(x -> next);
  printf(" v(%d)<-%d,%d ", x->var_id, x->vect1, x->vect2);
}

void pt(trans)
struct info_transition * trans;
{ 
  if (!trans) return;
  pt (trans-> next);
    printf("transition number:%d ", trans->transition_number);
    if (trans->synlabel)
      printf("DEST: %s/%d, SYNLAB: %s/%d, MATRIX-GUARD: %d\n",
        trans->destination, loc_index(trans->destination), trans->synlabel, 
        sync_index(trans->synlabel), trans->matrix_guard_id);
    else
      printf("DEST: %s/%d, NO-SYNLAB, MATRIX-GUARD: %d\n",
        trans->destination, loc_index(trans->destination),
        trans->matrix_guard_id);
    printf("---->assignments: ");
    print_assignments(trans->assign);
    if (trans->was_asap) printf("ASAP transition\n");
    printf("\n");
    trans=trans->next;
}

void print_transitions(trans)
struct info_transition * trans;
{
  pt(trans);
}

/* auxilliary for print_locations */
void pl(loc)
struct info_location * loc;
{ 
  int * in_trans, i;
 
  if (loc == NULL) return;
  pl(loc -> next);
  printf("\n******* LOC-NAME: %s/%d, GUARD-INV: %d\n",
   loc->location_name,loc->location_number,loc->matrix_inv_id);
  print_loc_slopes(loc);
  print_transitions(loc->transitions);/*first element is a dummy */
  printf("TRANSITIONS_TO_ME_FROM transitions: ");
  in_trans = transitions_to(loc->location_number);
  for (i = 0; in_trans[i] != -1; i++) printf(" %d", in_trans[i]);
  printf("\n");
  return;
}

void print_locations()
 { pl(locations); }







