/* routines related to predicates as they're parsed in (all in C) */



#include "predicate.h"

/* for storing last used matrix value in finding the spec constraints */
/* since constraints may come in fragmented parts */
int old_matrix_id;

struct convex_predicate *final_pred;
struct convex_predicate *convex_pred;
struct state_predicate *p_state_pred;

char *state_aut_name;


void
print_state_predicate(struct state_predicate *sp) {
  if (sp != NULL) {
    printf("\tAut: %s\n",sp->aut_str);
    printf("\tLoc: %s\n",sp->locn_str);
    print_state_predicate(sp->next);
  }
  ansi_fflush(stdout);
}

void
print_convex_predicate(struct convex_predicate *cp) {
  if (cp != NULL) {
    print_state_predicate(cp->state_pred);
    pm(cp->var_pred);
    print_convex_predicate(cp->next);
  }
  ansi_fflush(stdout);
}


void
init_predicate() {
  convex_pred = (struct convex_predicate *) NULL;
}

void
init_state_predicate() {
  p_state_pred = (struct state_predicate *) NULL;
}

void
init_convex_predicate() {
  init_state_predicate();
  /* need to store the  current matrix id, since the convex predicate
     may come in several segments, each appearing in a separate matrix */
  matrix_head = NULL; /* added 5/22 HWT */
  old_matrix_id = matrix_id;
  rows = 0;
}


/* insert it into the currently accumulated state predicate p_state_pred */
void
insert_state_predicate(char *aut_name, char *loc_name) {
  struct state_predicate *sp;
  sp = (struct state_predicate *) malloc (sizeof(struct state_predicate));
  sp->aut_str = str_preserve(aut_name);
  sp->locn_str = str_preserve(loc_name);
  sp->next = p_state_pred;
  p_state_pred = sp;
}

/* insert the currently accumulated convex predicate (found in globals */
/*   p_state_pred and vect_head(?)) in the final_global pred */
/* substantially revised, 5/22/95: HWT, removed bugs in joining matrix */
/* segments together into a single matrix */
void
assign_convex_spec() {
  struct convex_predicate *cp;  
  struct _p_matrix *mp, *mp2;
  struct _p_vector *vp, **insertvpaddr;
  struct _p_vector *vpnew;
  int i;

#ifdef __DEBUG_PRED
  fprintf(stdout,"Starting assign convex pred\n");
  ansi_fflush(stdout);
#endif

  cp = (struct convex_predicate *) malloc (sizeof(struct convex_predicate));
  cp->state_pred = p_state_pred;

#ifdef __DEBUG_PRED
  pm(matrix_head);
#endif

  /* now do the matrix */
  /*  join together all matrices from global list with id > old-matrix_id */
  /* assume the order is strictly decreasing */
  /* mp is ptr to the new matrix */
  mp = (struct _p_matrix *)malloc(sizeof(struct _p_matrix));
  mp2 = matrix_head;
  mp -> columns = columns;
  mp->rows = 0;
  mp->vectors = NULL;
  mp->next = NULL;
  insertvpaddr = & (mp->vectors);
  while ((mp2!= NULL) && (mp2->matrix_number > old_matrix_id)) {
    mp -> rows += mp2->rows-1;
    /* add the vector lists, excluding all final rows for now */
    vp = mp2->vectors;
    /* final row appears first, skip past it */
    vp = vp->next;
    while (vp != NULL) {
      /* make a new copy of the vectors */
      int i;
      struct _p_vector *vpnew = 
	(struct _p_vector *) malloc (sizeof(struct _p_vector));
      vpnew->len = vp->len; 
      vpnew->vect_id = vp->vect_id;
      vpnew->variable = vp->variable;
      vpnew->d = (int *) malloc (vp->len * sizeof(int));
      vpnew->n = (int *) malloc (vp->len * sizeof(int));
      for (i=0;i<vp->len;i++) {
	vpnew->d[i] = vp->d[i];
	vpnew->n[i] = vp->n[i];
      }
      /* insert at designated position, at tail */
      vpnew->next = NULL;
      *insertvpaddr = vpnew;
      insertvpaddr = &(vpnew->next);
      vp = vp->next;
    }
    mp2 = mp2->next;
  }

  /* just in case there are no matrices, and therefore no vectors at all */
  /* we must assign a final row */
  vpnew = 
	(struct _p_vector *) malloc (sizeof(struct _p_vector));
  vpnew->len = var_count+2;
  vpnew->vect_id = -1;
  vpnew->variable = -1; /* not an assignment */
  vpnew->d = (int *) malloc (vpnew->len * sizeof(int));
  vpnew->n = (int *) malloc (vpnew->len * sizeof(int));
  for (i=0;i<vpnew->len;i++) {
    vpnew->d[i] = 1;
    vpnew->n[i] = 0;
  }
  vpnew->n[0] = vpnew->n[vpnew->len-1] = 1;
  /* insert at designated position, at tail of matrix, i.e. head of list */
  vpnew->next = mp->vectors;
  mp->rows++;
  mp->vectors = vpnew;
  
  vect_head = NULL;
  cp->var_pred = mp;

  /* assign to list */
  cp->next = convex_pred;
  convex_pred = cp;

#ifdef __DEBUG_PRED
  print_convex_predicate(cp);
  fprintf(stdout,"Done with convex pred assign\n");
  ansi_fflush(stdout);
#endif
}

void
init_pred_spec() {
  init_predicate();
  /* are these the only initializations i need for the constraints? */
  vect_head = NULL;
  rows = 0;
}


void
assign_final_spec() {
  if (had_final_spec) {
    printf("Already declared final spec, redeclaration ignored\n");
  }
  else {
    /* it's actually a list of convex specs */
    final_pred = convex_pred;
    had_final_spec = 1;
  }
/*  print_convex_predicate(final_pred); */
}

